# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest

import affine
import numpy as np
import pyproj
import xarray as xr

from xcube_resampling.gridmapping import GridMapping
from xcube_resampling.gridmapping.coords import Coords2DGridMapping
from xcube_resampling.gridmapping.regular import RegularGridMapping

GEO_CRS = pyproj.crs.CRS(4326)
NOT_A_GEO_CRS = pyproj.crs.CRS(5243)


class _TestGridMapping(GridMapping):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.rgm = GridMapping.regular(
            size=self.size,
            tile_size=self.tile_size,
            is_j_axis_up=self.is_j_axis_up,
            xy_res=self.xy_res,
            xy_min=(self.xy_bbox[0] + self.x_res / 2, self.xy_bbox[1] + self.x_res / 2),
            crs=self.crs,
        )

    def _new_x_coords(self) -> xr.DataArray:
        return self.rgm.x_coords

    def _new_y_coords(self) -> xr.DataArray:
        return self.rgm.y_coords

    def _new_xy_coords(self) -> xr.DataArray:
        return self.rgm.xy_coords


# noinspection PyMethodMayBeStatic
class GridMappingTest(unittest.TestCase):
    _kwargs = dict(
        size=(720, 360),
        tile_size=(360, 180),
        xy_bbox=(-180.0, -90.0, 180.0, 90.0),
        xy_res=(360 / 720, 360 / 720),
        crs=GEO_CRS,
        xy_var_names=("x", "y"),
        xy_dim_names=("x", "y"),
        is_regular=True,
        is_lon_360=False,
        is_j_axis_up=False,
    )

    def kwargs(self, **kwargs):
        orig_kwargs = dict(self._kwargs)
        orig_kwargs.update(**kwargs)
        if "xy_min" in orig_kwargs:
            # Replace xy_min by xy_bbox.
            # Using xy_min instead of xy_bbox makes it easier for us
            width, height = orig_kwargs["size"]
            try:
                x_res, y_res = orig_kwargs["xy_res"]
            except TypeError:
                x_res, y_res = 2 * (orig_kwargs["xy_res"],)
            x_min, y_min = orig_kwargs.pop("xy_min")
            x_max, y_max = x_min + x_res * (width - 1), y_min + y_res * (height - 1)
            orig_kwargs["xy_bbox"] = (
                x_min - x_res / 2,
                y_min - y_res / 2,
                x_max + x_res / 2,
                y_max + y_res / 2,
            )
        return orig_kwargs

    def test_valid(self):
        gm = _TestGridMapping(**self.kwargs())
        self.assertEqual((720, 360), gm.size)
        self.assertEqual(720, gm.width)
        self.assertEqual(360, gm.height)
        self.assertEqual(True, gm.is_tiled)
        self.assertEqual((360, 180), gm.tile_size)
        self.assertEqual(360, gm.tile_width)
        self.assertEqual(180, gm.tile_height)
        self.assertEqual((0, 0, 720, 360), gm.ij_bbox)
        self.assertEqual((-180.0, -90.0, 180.0, 90.0), gm.xy_bbox)
        self.assertEqual(-179.75, gm.x_min)
        self.assertEqual(-89.75, gm.y_min)
        self.assertEqual(179.75, gm.x_max)
        self.assertEqual(89.75, gm.y_max)
        self.assertEqual((0.5, 0.5), gm.xy_res)
        self.assertEqual(0.5, gm.x_res)
        self.assertEqual(0.5, gm.y_res)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual("degree", gm.spatial_unit_name)
        self.assertEqual(True, gm.is_regular)
        self.assertEqual(False, gm.is_lon_360)
        self.assertEqual(False, gm.is_j_axis_up)

        self.assertIsInstance(gm.xy_coords, xr.DataArray)
        np.testing.assert_equal(
            gm.ij_bboxes,
            np.array(
                [
                    [0, 0, 360, 180],
                    [360, 0, 720, 180],
                    [0, 180, 360, 360],
                    [360, 180, 720, 360],
                ]
            ),
        )
        np.testing.assert_equal(
            gm.xy_bboxes,
            np.array(
                [
                    [-180.0, 0.0, 0.0, 90.0],
                    [0.0, 0.0, 180.0, 90.0],
                    [-180.0, -90.0, 0.0, 0.0],
                    [0.0, -90.0, 180.0, 0.0],
                ]
            ),
        )

    def test_invalids(self):
        with self.assertRaises(ValueError) as cm:
            _TestGridMapping(**self.kwargs(size=(360, 1)))
        self.assertEqual("invalid size", f"{cm.exception}")

        with self.assertRaises(ValueError) as cm:
            _TestGridMapping(**self.kwargs(size=(360,)))
        self.assertEqual(
            "not enough values to unpack (expected 2, got 1)", f"{cm.exception}"
        )

        with self.assertRaises(ValueError) as cm:
            _TestGridMapping(**self.kwargs(size=None))
        self.assertEqual(
            "size must be an int or a sequence of two ints", f"{cm.exception}"
        )

        with self.assertRaises(ValueError) as cm:
            _TestGridMapping(**self.kwargs(tile_size=0))
        self.assertEqual("invalid tile_size", f"{cm.exception}")

        with self.assertRaises(ValueError) as cm:
            _TestGridMapping(**self.kwargs(xy_res=-0.1))
        self.assertEqual("invalid xy_res", f"{cm.exception}")

    def test_scalars(self):
        gm = _TestGridMapping(**self.kwargs(size=360, tile_size=180, xy_res=0.1))
        self.assertEqual((360, 360), gm.size)
        self.assertEqual((180, 180), gm.tile_size)
        self.assertEqual((0.1, 0.1), gm.xy_res)

    def test_not_tiled(self):
        gm = _TestGridMapping(**self.kwargs(tile_size=None))
        self.assertEqual((720, 360), gm.tile_size)
        self.assertEqual(False, gm.is_tiled)

    def test_ij_to_xy_transform(self):
        image_geom = _TestGridMapping(
            **self.kwargs(size=(1200, 1200), xy_min=(0, 0), xy_res=1, crs=NOT_A_GEO_CRS)
        )
        i2crs = image_geom.ij_to_xy_transform
        self.assertMatrixPoint((0, 0), i2crs, (0, 1199))
        self.assertMatrixPoint((1024, 0), i2crs, (1024, 1199))
        self.assertMatrixPoint((0, 1024), i2crs, (0, 1199 - 1024))
        self.assertMatrixPoint((1024, 1024), i2crs, (1024, 1199 - 1024))
        self.assertEqual(((1, 0, 0), (0.0, -1, 1199)), i2crs)

        image_geom = _TestGridMapping(
            **self.kwargs(size=(1440, 720), xy_min=(-179.875, -89.875), xy_res=0.25)
        )
        i2crs = image_geom.ij_to_xy_transform
        self.assertMatrixPoint((-179.875, 89.875), i2crs, (0, 0))
        self.assertMatrixPoint((0.125, -0.125), i2crs, (720, 360))
        self.assertMatrixPoint((179.875, -89.875), i2crs, (1439, 719))
        self.assertEqual(((0.25, 0.0, -179.875), (0.0, -0.25, 89.875)), i2crs)

        image_geom = _TestGridMapping(
            **self.kwargs(
                size=(1440, 720),
                xy_min=(-179.875, -89.875),
                xy_res=0.25,
                is_j_axis_up=True,
            )
        )
        i2crs = image_geom.ij_to_xy_transform
        self.assertMatrixPoint((-179.875, -89.875), i2crs, (0, 0))
        self.assertMatrixPoint((0.125, 0.125), i2crs, (720, 360))
        self.assertMatrixPoint((179.875, 89.875), i2crs, (1439, 719))
        self.assertEqual(((0.25, 0.0, -179.875), (0.0, 0.25, -89.875)), i2crs)

    def test_xy_to_ij_transform(self):
        image_geom = _TestGridMapping(
            **self.kwargs(size=(1200, 1200), xy_min=(0, 0), xy_res=1, crs=NOT_A_GEO_CRS)
        )
        crs2i = image_geom.xy_to_ij_transform
        self.assertMatrixPoint((0, 0), crs2i, (0, 1199))
        self.assertMatrixPoint((1024, 0), crs2i, (1024, 1199))
        self.assertMatrixPoint((0, 1024), crs2i, (0, 1199 - 1024))
        self.assertMatrixPoint((1024, 1024), crs2i, (1024, 1199 - 1024))
        self.assertEqual(((1, 0, 0), (0.0, -1, 1199)), crs2i)

        image_geom = _TestGridMapping(**self.kwargs(size=(1440, 720), xy_res=0.25))
        crs2i = image_geom.xy_to_ij_transform
        self.assertMatrixPoint((0, 719), crs2i, (-179.875, -89.875))
        self.assertMatrixPoint((720, 360), crs2i, (0.125, -0.125))
        self.assertMatrixPoint((1439, 0), crs2i, (179.875, 89.875))
        self.assertEqual(((4.0, 0.0, 719.5), (0.0, -4.0, 359.5)), crs2i)

        image_geom = _TestGridMapping(
            **self.kwargs(size=(1440, 720), xy_res=0.25, is_j_axis_up=True)
        )
        crs2i = image_geom.xy_to_ij_transform
        self.assertMatrixPoint((0, 0), crs2i, (-179.875, -89.875))
        self.assertMatrixPoint((720, 360), crs2i, (0.125, 0.125))
        self.assertMatrixPoint((1439, 719), crs2i, (179.875, 89.875))
        self.assertEqual(((4.0, -0.0, 719.5), (-0.0, 4.0, 359.5)), crs2i)

    def test_ij_transform_to_and_from(self):
        gm1 = _TestGridMapping(
            **self.kwargs(size=(1440, 720), xy_res=0.25, is_j_axis_up=True)
        )
        gm2 = _TestGridMapping(
            **self.kwargs(
                size=(1000, 1000), xy_min=(10, 50), xy_res=0.025, is_j_axis_up=True
            )
        )
        self.assertEqual(
            ((10.0, 0.0, -7595.0), (0.0, 10.0, -5595.0)), gm1.ij_transform_to(gm2)
        )
        self.assertEqual(
            ((10.0, -0.0, -7595.0), (-0.0, 10.0, -5595.0)), gm2.ij_transform_from(gm1)
        )
        self.assertEqual(
            ((0.1, 0.0, 759.5), (0.0, 0.1, 559.5)), gm2.ij_transform_to(gm1)
        )
        self.assertEqual(
            ((0.1, 0.0, 759.5), (0.0, 0.1, 559.5)), gm1.ij_transform_from(gm2)
        )

    def assertMatrixPoint(self, expected_point, matrix, point):
        affine_matrix = affine.Affine(*matrix[0], *matrix[1])
        actual_point = affine_matrix * point
        self.assertAlmostEqual(expected_point[0], actual_point[0])
        self.assertAlmostEqual(expected_point[1], actual_point[1])
        return actual_point

    def test_derive(self):
        gm = _TestGridMapping(**self.kwargs())
        self.assertEqual((720, 360), gm.size)
        self.assertEqual((360, 180), gm.tile_size)
        self.assertEqual(False, gm.is_j_axis_up)

        # force creating of xy_coords array and save value
        xy_coords = gm.xy_coords

        derived_gm = gm.derive(
            tile_size=270,
            is_j_axis_up=True,
            xy_var_names=("u", "v"),
            xy_dim_names=("i", "j"),
        )

        self.assertIsNot(gm, derived_gm)
        self.assertIsInstance(derived_gm, _TestGridMapping)
        self.assertEqual((720, 360), derived_gm.size)
        self.assertEqual((270, 270), derived_gm.tile_size)
        self.assertEqual(True, derived_gm.is_j_axis_up)
        self.assertEqual(("u", "v"), derived_gm.xy_var_names)
        self.assertEqual(("i", "j"), derived_gm.xy_dim_names)

        derived_xy_coords = derived_gm.xy_coords
        self.assertIsNot(xy_coords, derived_xy_coords)
        self.assertEqual(((2,), (270, 90), (270, 270, 180)), derived_xy_coords.chunks)

    def test_scale(self):
        gm = _TestGridMapping(**self.kwargs())
        self.assertEqual((720, 360), gm.size)
        self.assertEqual((360, 180), gm.tile_size)
        self.assertEqual(False, gm.is_j_axis_up)

        # force creating of xy_coords array and save value
        xy_coords = gm.xy_coords

        scaled_gm = gm.scale((0.25, 0.5))

        self.assertIsNot(gm, scaled_gm)
        self.assertIsInstance(scaled_gm, RegularGridMapping)
        self.assertEqual((180, 180), scaled_gm.size)
        self.assertEqual((180, 180), scaled_gm.tile_size)
        self.assertEqual(False, scaled_gm.is_j_axis_up)
        self.assertEqual(("x", "y"), scaled_gm.xy_var_names)
        self.assertEqual(("x", "y"), scaled_gm.xy_dim_names)

        scaled_xy_coords = scaled_gm.xy_coords
        self.assertIsNot(xy_coords, scaled_xy_coords)
        self.assertEqual(((2,), (180,), (180,)), scaled_xy_coords.chunks)

        scaled_gm = gm.scale((0.25, 0.5), tile_size=(90, 90))

        self.assertIsNot(gm, scaled_gm)
        self.assertIsInstance(scaled_gm, RegularGridMapping)
        self.assertEqual((180, 180), scaled_gm.size)
        self.assertEqual((90, 90), scaled_gm.tile_size)
        self.assertEqual(False, scaled_gm.is_j_axis_up)
        self.assertEqual(("x", "y"), scaled_gm.xy_var_names)
        self.assertEqual(("x", "y"), scaled_gm.xy_dim_names)

        scaled_xy_coords = scaled_gm.xy_coords
        self.assertIsNot(xy_coords, scaled_xy_coords)
        self.assertEqual(((2,), (90, 90), (90, 90)), scaled_xy_coords.chunks)

    def test_transform(self):
        gm = _TestGridMapping(
            **self.kwargs(
                xy_min=(20, 56),
                size=(400, 200),
                tile_size=(400, 200),
                xy_res=(0.01, 0.01),
            )
        )
        transformed_gm = gm.transform("EPSG:32633")

        self.assertIsNot(gm, transformed_gm)
        self.assertIsInstance(transformed_gm, Coords2DGridMapping)
        self.assertEqual(pyproj.CRS.from_string("EPSG:32633"), transformed_gm.crs)
        self.assertEqual((400, 200), transformed_gm.size)
        self.assertEqual((400, 200), transformed_gm.tile_size)
        self.assertEqual(False, transformed_gm.is_j_axis_up)
        self.assertEqual(
            ("transformed_x", "transformed_y"), transformed_gm.xy_var_names
        )
        self.assertEqual(("x", "y"), transformed_gm.xy_dim_names)

    def test_to_regular(self):
        gm = _TestGridMapping(
            **self.kwargs(
                xy_min=(9.6, 47.6),
                size=(1000, 1000),
                tile_size=(1000, 1000),
                xy_res=(0.0002, 0.0002),
            )
        )
        transformed_gm = gm.transform("EPSG:32633")
        transformed_gm_regular = transformed_gm.to_regular()

        self.assertIsNot(gm, transformed_gm_regular)
        self.assertIsInstance(transformed_gm_regular, RegularGridMapping)
        self.assertEqual(
            pyproj.CRS.from_string("EPSG:32633"), transformed_gm_regular.crs
        )
        self.assertEqual((1102, 1046), transformed_gm_regular.size)
        self.assertEqual((1000, 1000), transformed_gm_regular.tile_size)
        self.assertEqual(False, transformed_gm_regular.is_j_axis_up)
        self.assertEqual(False, transformed_gm_regular.is_lon_360)

    def test_is_close(self):
        gm1 = _TestGridMapping(
            **self.kwargs(xy_min=(0, 0), size=(400, 200), xy_res=(0.01, 0.01))
        )
        gm2 = _TestGridMapping(
            **self.kwargs(xy_min=(0, 0), size=(400, 200), xy_res=(0.01, 0.01))
        )
        self.assertTrue(gm1.is_close(gm1))
        self.assertTrue(gm2.is_close(gm2))
        self.assertTrue(gm1.is_close(gm2))
        self.assertTrue(gm2.is_close(gm1))

        tolerance = 0.001

        gm2 = _TestGridMapping(
            **self.kwargs(
                xy_min=(tolerance / 2, tolerance / 2),
                size=(400, 200),
                xy_res=(0.01, 0.01),
            )
        )
        self.assertTrue(gm1.is_close(gm1, tolerance=tolerance))
        self.assertTrue(gm2.is_close(gm2, tolerance=tolerance))
        self.assertTrue(gm1.is_close(gm2, tolerance=tolerance))
        self.assertTrue(gm2.is_close(gm1, tolerance=tolerance))

        gm2 = _TestGridMapping(
            **self.kwargs(
                xy_min=(tolerance * 2, tolerance * 2),
                size=(400, 200),
                xy_res=(0.01, 0.01),
            )
        )
        self.assertTrue(gm1.is_close(gm1, tolerance=tolerance))
        self.assertTrue(gm2.is_close(gm2, tolerance=tolerance))
        self.assertFalse(gm1.is_close(gm2, tolerance=tolerance))
        self.assertFalse(gm2.is_close(gm1, tolerance=tolerance))

    def test_is_regular_raise_error(self):
        gm = _TestGridMapping(**self.kwargs(is_regular=False))
        with self.assertRaises(ValueError) as cm:
            GridMapping.assert_regular(gm)
        self.assertIn("must be a regular grid mapping", str(cm.exception))

        with self.assertRaises(NotImplementedError) as cm:
            gm._assert_regular()
        self.assertIn(
            "Operation not implemented for non-regular grid mappings", str(cm.exception)
        )

    def test_repr_markdown(self):
        gm = _TestGridMapping(**self.kwargs())
        md = gm._repr_markdown_()

        # Basic structure
        self.assertIn("class: **_TestGridMapping**", md)
        self.assertIn("* is_regular: True", md)
        self.assertIn("* is_j_axis_up: False", md)
        self.assertIn("* is_lon_360: False", md)
        self.assertIn("* crs: EPSG:4326", md)
        self.assertIn("* xy_res: (0.5, 0.5)", md)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import unittest

import numpy as np

from xcube_resampling.gridmapping import CRS_WGS84, GridMapping
from xcube_resampling.reproject import reproject_dataset

from .sampledata import (
    create_2x5x5_dataset_regular_utm,
    create_5x5_dataset_regular_utm,
    create_5x5_dataset_regular_utm_antimeridian,
    create_large_dataset_for_reproject,
)


# noinspection PyMethodMayBeStatic
class ReprojectDatasetTest(unittest.TestCase):
    def test_reproject_target_gm(self):
        source_ds = create_5x5_dataset_regular_utm()

        # test projected CRS, similar resolution
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(4320120, 3382520), xy_res=80, crs="epsg:3035"
        )
        target_ds = reproject_dataset(source_ds, target_gm)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [1, 1, 2, 3, 4],
                    [6, 6, 7, 8, 9],
                    [11, 12, 12, 13, 14],
                    [16, 17, 17, 18, 19],
                    [21, 17, 17, 18, 19],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

    def test_reproject_target_gm_3d(self):
        source_ds = create_2x5x5_dataset_regular_utm()

        # test projected CRS, similar resolution
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(4320120, 3382520), xy_res=80, crs="epsg:3035"
        )
        target_ds = reproject_dataset(source_ds, target_gm)
        self.assertEqual(
            set(source_ds.variables),
            set(target_ds.variables),
        )
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [
                        [1, 1, 2, 3, 4],
                        [6, 6, 7, 8, 9],
                        [11, 12, 12, 13, 14],
                        [16, 17, 17, 18, 19],
                        [21, 17, 17, 18, 19],
                    ],
                    [
                        [1, 1, 2, 3, 4],
                        [6, 6, 7, 8, 9],
                        [11, 12, 12, 13, 14],
                        [16, 17, 17, 18, 19],
                        [21, 17, 17, 18, 19],
                    ],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

    def test_reproject_target_gm_j_axis_up(self):
        source_ds = create_5x5_dataset_regular_utm()
        target_gm = GridMapping.regular(
            size=(5, 5),
            xy_min=(4320120, 3382520),
            xy_res=80,
            crs="epsg:3035",
            is_j_axis_up=True,
        )
        target_ds = reproject_dataset(source_ds, target_gm)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [21, 17, 17, 18, 19],
                    [16, 17, 17, 18, 19],
                    [11, 12, 12, 13, 14],
                    [6, 6, 7, 8, 9],
                    [1, 1, 2, 3, 4],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

    def test_reproject_source_gm_j_axis_up(self):
        source_ds = create_5x5_dataset_regular_utm()
        source_ds = source_ds.isel(y=slice(None, None, -1))
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(4320120, 3382520), xy_res=80, crs="epsg:3035"
        )
        target_ds = reproject_dataset(source_ds, target_gm)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [1, 1, 2, 3, 4],
                    [6, 6, 7, 8, 9],
                    [11, 12, 12, 13, 14],
                    [16, 17, 17, 18, 19],
                    [21, 17, 17, 18, 19],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

    def test_reproject_target_gm_finer_res(self):
        source_ds = create_5x5_dataset_regular_utm()
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(4320090, 3382490), xy_res=20, crs="epsg:3035"
        )
        target_ds = reproject_dataset(source_ds, target_gm)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [15, 16, 16, 16, 16],
                    [15, 16, 16, 16, 16],
                    [15, 16, 16, 16, 16],
                    [20, 21, 21, 21, 21],
                    [20, 21, 21, 21, 21],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

    def test_reproject_target_gm_coarser_res(self):
        source_ds = create_5x5_dataset_regular_utm()
        target_gm = GridMapping.regular(
            size=(3, 3), xy_min=(4320070, 3382620), xy_res=120, crs="epsg:3035"
        )
        target_ds = reproject_dataset(source_ds, target_gm)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [0, 1, 3],
                    [5, 6, 8],
                    [15, 11, 13],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

    def test_reproject_target_gm_geographic_crs(self):
        source_ds = create_5x5_dataset_regular_utm()
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(9.9889, 53.5502), xy_res=0.0006, crs=CRS_WGS84
        )
        target_ds = reproject_dataset(source_ds, target_gm)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [7, 8, 8, 8, 9],
                    [12, 13, 13, 13, 14],
                    [12, 13, 13, 13, 14],
                    [17, 18, 18, 18, 19],
                    [22, 23, 23, 23, 24],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

    def test_reproject_target_gm_geographic_crs_antimeridian(self):
        source_ds = create_5x5_dataset_regular_utm_antimeridian()

        # Geographic target grid placed near 179.99°E
        target_gm = GridMapping.regular(
            size=(5, 5),
            xy_min=(179.998, 9.95),
            xy_res=0.001,
            crs=CRS_WGS84,
        )

        fill_value = -9999
        target_ds = reproject_dataset(
            source_ds,
            target_gm,
            fill_values=fill_value,
        )

        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [-9999, -9999, -9999, -9999, -9999],
                    [0, 1, 2, 3, 4],
                    [5, 6, 7, 8, 9],
                    [10, 11, 12, 13, 14],
                    [15, 16, 17, 18, 19],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )
        np.testing.assert_almost_equal(
            target_ds.lon.values,
            np.array(
                [179.998, 179.999, 180.0, -179.999, -179.998],
                dtype=target_ds.lon.dtype,
            ),
        )

    def test_reproject_target_gm_geographic_crs_fine_res(self):
        source_ds = create_5x5_dataset_regular_utm()

        # test geographic CRS with 1/2 resolution
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(9.98875, 53.55005), xy_res=0.0003, crs=CRS_WGS84
        )
        target_ds = reproject_dataset(source_ds, target_gm)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.array(
                [
                    [12, 12, 12, 13, 13],
                    [17, 17, 17, 18, 18],
                    [17, 17, 17, 18, 18],
                    [22, 17, 17, 18, 18],
                    [22, 22, 22, 23, 23],
                ],
                dtype=target_ds.band_1.dtype,
            ),
        )

    def test_reproject_complex_dask_array(self):
        source_ds = create_large_dataset_for_reproject()
        target_gm = GridMapping.regular(
            size=(10, 10),
            xy_min=(6.1, 48.1),
            xy_res=0.2,
            crs=CRS_WGS84,
            tile_size=(5, 5),
        )

        target_ds = reproject_dataset(source_ds, target_gm, interp_methods="triangular")
        self.assertCountEqual(["temperature", "onedim_data"], list(target_ds.data_vars))
        self.assertAlmostEqual(target_ds.temperature.values[0, 0, 0], 6216.14, places=2)
        self.assertAlmostEqual(
            target_ds.temperature.values[0, -1, -1], 2848.50, places=2
        )
        self.assertEqual(
            [2, 5, 5],
            [
                target_ds.temperature.chunksizes["time"][0],
                target_ds.temperature.chunksizes["lat"][0],
                target_ds.temperature.chunksizes["lon"][0],
            ],
        )

        target_ds = reproject_dataset(source_ds, target_gm, interp_methods=1)
        self.assertCountEqual(["temperature", "onedim_data"], list(target_ds.data_vars))
        self.assertAlmostEqual(target_ds.temperature.values[0, 0, 0], 6216.14, places=2)
        self.assertAlmostEqual(
            target_ds.temperature.values[0, -1, -1], 2848.50, places=2
        )
        self.assertEqual(
            [2, 5, 5],
            [
                target_ds.temperature.chunksizes["time"][0],
                target_ds.temperature.chunksizes["lat"][0],
                target_ds.temperature.chunksizes["lon"][0],
            ],
        )

    def test_reproject_raise_not_implemented(self):
        source_ds = create_5x5_dataset_regular_utm()
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(4320080, 3382480), xy_res=20, crs="epsg:3035"
        )
        with self.assertRaises(NotImplementedError) as cm:
            _ = reproject_dataset(source_ds, target_gm, interp_methods="cubic")
        self.assertIn(
            "interp_methods must be one of 0, 1, 'nearest', 'bilinear', 'triangular'",
            str(cm.exception),
        )

    def test_reproject_outside_return_empty_ds(self):
        source_ds = create_5x5_dataset_regular_utm()
        target_gm = GridMapping.regular(
            size=(5, 5), xy_min=(4320550, 3382920), xy_res=80, crs="epsg:3035"
        )
        target_ds = reproject_dataset(source_ds, target_gm)
        np.testing.assert_almost_equal(
            target_ds.band_1.values,
            np.full((5, 5), -1, dtype=target_ds.band_1.dtype),
        )



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest

import numpy as np
import pyproj
import xarray as xr

from xcube_resampling.gridmapping import GridMapping

from ..sampledata import create_s2plus_dataset

GEO_CRS = pyproj.crs.CRS(4326)
NOT_A_GEO_CRS = pyproj.crs.CRS(5243)


# noinspection PyMethodMayBeStatic
class DatasetGridMappingTest(unittest.TestCase):
    def test_from_non_regular_cube(self):
        lon = np.array(
            [[8, 9.3, 10.6, 11.9], [8, 9.2, 10.4, 11.6], [8, 9.1, 10.2, 11.3]],
            dtype=np.float32,
        )
        lat = np.array(
            [[56, 56.1, 56.2, 56.3], [55, 55.2, 55.4, 55.6], [54, 54.3, 54.6, 54.9]],
            dtype=np.float32,
        )
        rad = np.random.random(3 * 4).reshape((3, 4))
        dims = ("y", "x")
        dataset = xr.Dataset(
            dict(
                lon=xr.DataArray(lon, dims=dims),
                lat=xr.DataArray(lat, dims=dims),
                rad=xr.DataArray(rad, dims=dims),
            )
        )
        gm = GridMapping.from_dataset(dataset)
        self.assertEqual((4, 3), gm.size)
        self.assertEqual((4, 3), gm.tile_size)
        self.assertEqual(GEO_CRS, gm.crs)
        self.assertEqual(False, gm.is_regular)
        self.assertEqual(False, gm.is_lon_360)
        self.assertEqual(False, gm.is_j_axis_up)
        self.assertEqual((2, 3, 4), gm.xy_coords.shape)
        self.assertEqual(("coord", "y", "x"), gm.xy_coords.dims)
        self.assertEqual((1.2, 0.85), gm.xy_res)

    def test_crs(self):
        ds = xr.Dataset(
            {
                "var": (("lat", "lon"), np.random.rand(2, 2)),
            },
            coords={
                "lon": ("lon", [0, 1]),
                "lat": ("lat", [0, 1]),
            },
        )

        result = GridMapping.from_dataset(ds, crs="EPSG:4326")
        self.assertTrue(result.is_regular)
        self.assertEqual(result.crs.to_string(), "EPSG:4326")

    def test_from_sentinel_2(self):
        dataset = create_s2plus_dataset()
        tol = 1e-6

        gm = GridMapping.from_dataset(dataset, tolerance=tol)
        # Should pick the projected one which is regular
        self.assertIn("Projected", gm.crs.type_name)
        self.assertEqual(True, gm.is_regular)

        gm = GridMapping.from_dataset(dataset, prefer_is_regular=True, tolerance=tol)
        # Should pick the projected one which is regular
        self.assertIn("Projected", gm.crs.type_name)
        self.assertEqual(True, gm.is_regular)

        gm = GridMapping.from_dataset(dataset, prefer_is_regular=False, tolerance=tol)
        # Should pick the geographic one which is irregular
        self.assertIn("Geographic", gm.crs.type_name)
        self.assertEqual(False, gm.is_regular)

        gm = GridMapping.from_dataset(dataset, prefer_crs=GEO_CRS, tolerance=tol)
        # Should pick the geographic one which is irregular
        self.assertIn("Geographic", gm.crs.type_name)
        self.assertEqual(False, gm.is_regular)

        gm = GridMapping.from_dataset(
            dataset, prefer_crs=GEO_CRS, prefer_is_regular=True, tolerance=tol
        )
        # Should pick the geographic one which is irregular
        self.assertIn("Geographic", gm.crs.type_name)
        self.assertEqual(False, gm.is_regular)

    def test_no_grid_mapping_found(self):
        with self.assertRaises(ValueError) as cm:
            GridMapping.from_dataset(xr.Dataset())
        self.assertEqual("cannot find any grid mapping in dataset", f"{cm.exception}")



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""This module comprises reducer functions to be passed to
`dask.array.coarsen()` that either do not exist in numpy, like `mode`,
or have special implementations compared to their numpy equivalents.
"""

import unittest

import numpy as np

from xcube_resampling import coarsen


class TestCoarsenAllInOne(unittest.TestCase):

    def test_all_reducers(self):
        # Prepare test arrays
        arr_float = np.array([[1.0, 2.0], [3.0, 4.0]])
        arr_int = np.array([[1, 2], [3, 4]])
        arr_mode = np.array([[1, 2, 2], [3, 2, 2]])

        axis = (0, 1)

        # ---- Reducers with specified axis ----
        np.testing.assert_array_equal(coarsen.first(arr_float, axis), np.array(1.0))
        np.testing.assert_array_equal(coarsen.last(arr_float, axis), np.array(4.0))
        np.testing.assert_array_equal(coarsen.center(arr_float, axis), np.array(4.0))
        np.testing.assert_array_equal(coarsen.mean(arr_float, axis), np.array(2.5))
        np.testing.assert_array_equal(coarsen.mean(arr_int, axis), np.array(2))
        np.testing.assert_array_equal(coarsen.median(arr_float, axis), np.array(2.5))
        np.testing.assert_array_almost_equal(
            coarsen.std(arr_float, axis), np.array(np.std(arr_float))
        )
        np.testing.assert_array_equal(coarsen.sum(arr_int, axis), np.array(10))
        np.testing.assert_array_almost_equal(
            coarsen.var(arr_float, axis), np.array(np.var(arr_float))
        )
        np.testing.assert_array_equal(coarsen.mode(arr_mode, axis), np.array(2))

        # ---- Reducers with axis=None (edge block pass-through) ----
        for reducer in [coarsen.first, coarsen.last, coarsen.center, coarsen.mode]:
            np.testing.assert_array_equal(reducer(arr_float, axis=None), arr_float)



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


import logging
from collections.abc import Callable, Hashable, Mapping
from typing import Annotated, Literal, Sequence, TypeAlias

import numpy as np

from .coarsen import center, first, last, mean, median, mode, std, var

__all__ = [
    "FloatInt",
    "AffineTransformMatrix",
    "SpatialAggMethod",
    "SpatialAggMethods",
    "SpatialInterpMethodInt",
    "SpatialInterpMethodStr",
    "SpatialInterpMethod",
    "SpatialInterpMethods",
    "TemporalAggMethod",
    "TemporalAggMethods",
    "TemporalInterpMethods",
    "TemporalInterpMethod",
    "PreventNaNPropagations",
    "FillValues",
    "FILLVALUE_UINT8",
    "FILLVALUE_UINT16",
    "FILLVALUE_INT",
    "FILLVALUE_FLOAT",
    "LOG",
]

FloatInt = int | float
"""A type alias representing either a float or an int."""

AffineTransformMatrix = tuple[
    tuple[FloatInt, FloatInt, FloatInt], tuple[FloatInt, FloatInt, FloatInt]
]
"""A 2×3 affine transformation matrix represented as nested tuples."""

SpatialAggMethod: TypeAlias = Literal[
    "center",
    "count",
    "first",
    "last",
    "max",
    "mean",
    "median",
    "mode",
    "min",
    "prod",
    "std",
    "sum",
    "var",
]
"""A literal type representing supported spatial aggregation methods."""

SpatialAggMethods: TypeAlias = (
    SpatialAggMethod | Mapping[np.dtype | str, SpatialAggMethod]
)
"""An spatial aggregation method or a mapping from variable name or dtype to 
spatial aggregation method.
"""

SpatialInterpMethodInt = Literal[0, 1]
"""Spatial interpolation method, as integer code."""
SpatialInterpMethodStr = Literal["nearest", "triangular", "bilinear"]
"""Spatial interpolation method, as string literal."""
SpatialInterpMethod = SpatialInterpMethodInt | SpatialInterpMethodStr
"""Spatial interpolation method, as integer code or string literal."""
SpatialInterpMethods: TypeAlias = (
    SpatialInterpMethod | Mapping[np.dtype | Hashable, SpatialInterpMethod]
)
"""A spatial interpolation method or a mapping from variable name or dtype to 
interpolation method.
"""

PercentileString: TypeAlias = Annotated[str, "format: percentile_<int>"]
TemporalAggMethod = (
    Literal[
        "all",
        "any",
        "backfill",
        "bfill",
        "count",
        "cumprod",
        "cumsum",
        "ffill",
        "first",
        "last",
        "max",
        "min",
        "mean",
        "median",
        "nearest",
        "pad",
        "prod",
        "std",
        "sum",
        "var",
    ]
    | PercentileString
)
"""A literal type representing supported temporal aggregation methods."""
TemporalAggMethods: TypeAlias = (
    TemporalAggMethod
    | Sequence[TemporalAggMethod]
    | Mapping[np.dtype | str, TemporalAggMethod | Sequence[TemporalAggMethod]]
)
"""An temporal aggregation method, a list of temporal aggregation methods, or a 
mapping from variable name or dtype to temporal aggregation method(s).
"""

TemporalInterpMethod = Literal[
    "linear", "nearest", "zero", "slinear", "quadratic", "cubic", "polynomial"
]
"""Temporal interpolation method, as string literal."""
TemporalInterpMethods: TypeAlias = (
    TemporalInterpMethod
    | Sequence[TemporalInterpMethod]
    | Mapping[
        np.dtype | Hashable, TemporalInterpMethod | Sequence[TemporalInterpMethod]
    ]
)
"""A temporal interpolation method, a list of temporal interpolation methods, or a 
mapping from variable name or dtype to interpolation method(s).
"""

PreventNaNPropagations: TypeAlias = bool | Mapping[np.dtype | str, bool]
"""It True, it prevents NaN propagation during upsampling (only applies when 
interpolation method is not nearest). It can be set globally or per data variable
or dtype.
"""

FillValues: TypeAlias = FloatInt | Mapping[np.dtype | str, FloatInt]
"""Fill values for missing data, as a scalar or mapping from dtype to value."""

FILLVALUE_UINT8 = 255
"""Default fill value for uint8 arrays."""

FILLVALUE_UINT16 = 65535
"""Default fill value for uint16 arrays."""

FILLVALUE_UINT32 = 4294967295
"""Default fill value for uint32 arrays."""

FILLVALUE_INT = -1
"""Default fill value for integer arrays."""

FILLVALUE_FLOAT = np.nan
"""Default fill value for floating-point arrays."""

# Internal helpers (not exported in __all__)
AggFunction: TypeAlias = Callable[[np.ndarray, tuple[int, ...] | None], np.ndarray]
AGG_METHODS: dict[SpatialAggMethod, AggFunction] = {
    "center": center,
    "count": np.count_nonzero,
    "first": first,
    "last": last,
    "prod": np.nanprod,
    "max": np.nanmax,
    "mean": mean,
    "median": median,
    "min": np.nanmin,
    "mode": mode,
    "std": std,
    "sum": np.nansum,
    "var": var,
}
INTERP_METHOD_MAPPING = {0: "nearest", 1: "bilinear", "nearest": 0, "bilinear": 1}

SCALE_LIMIT = 0.95
UV_DELTA = 1e-3

LOG = logging.getLogger("xcube.resampling")



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import dask.array as da
import numpy as np
import pyproj
import xarray as xr

from xcube_resampling.constants import LOG

from .base import DEFAULT_TOLERANCE, GridMapping
from .coords import new_grid_mapping_from_coords
from .helpers import (
    _assert_valid_xy_names,
    _normalize_crs,
)


def transform_grid_mapping(
    grid_mapping: GridMapping,
    crs: str | pyproj.crs.CRS,
    *,
    tile_size: int | tuple[int, int] = None,
    xy_var_names: tuple[str, str] = None,
    tolerance: float = DEFAULT_TOLERANCE,
) -> GridMapping:
    target_crs = _normalize_crs(crs)
    if xy_var_names:
        _assert_valid_xy_names(xy_var_names, name="xy_var_names")

    source_crs = grid_mapping.crs
    if source_crs == target_crs:
        LOG.warning(
            "Skipping grid mapping transformation: source and target CRS "
            "are identical (%s).",
            source_crs,
        )
        if tile_size is not None or xy_var_names is not None:
            return grid_mapping.derive(tile_size=tile_size, xy_var_names=xy_var_names)
        return grid_mapping

    if grid_mapping.x_coords.ndim == 1:
        x = da.asarray(grid_mapping.x_coords.data).rechunk(grid_mapping.tile_width)
        y = da.asarray(grid_mapping.y_coords.data).rechunk(grid_mapping.tile_height)
        xx, yy = da.meshgrid(x, y)
    else:
        xx = da.asarray(grid_mapping.x_coords.data)
        yy = da.asarray(grid_mapping.y_coords.data)

    transformer = pyproj.transformer.Transformer.from_crs(
        source_crs, target_crs, always_xy=True
    )

    def transform_block(xx: np.ndarray, yy: np.ndarray) -> np.ndarray:
        x, y = transformer.transform(xx, yy)
        return np.stack([x, y])

    xy_transformed = da.map_blocks(
        transform_block,
        xx,
        yy,
        dtype=np.float32,
        chunks=(2, *xx.chunks),
    )
    xx_transformed = xy_transformed[0]
    yy_transformed = xy_transformed[1]

    xy_var_names = xy_var_names or ("transformed_x", "transformed_y")

    if tile_size is None:
        tile_size = grid_mapping.tile_size

    yx_dim = (grid_mapping.xy_dim_names[1], grid_mapping.xy_dim_names[0])
    return new_grid_mapping_from_coords(
        x_coords=xr.DataArray(xx_transformed, dims=yx_dim, name=xy_var_names[0]),
        y_coords=xr.DataArray(yy_transformed, dims=yx_dim, name=xy_var_names[1]),
        crs=target_crs,
        tile_size=tile_size,
        tolerance=tolerance,
    )



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import math
from collections.abc import Hashable, Sequence
from fractions import Fraction
from typing import Any

import affine
import dask.array as da
import numpy as np
import pyproj.crs
import xarray as xr

from xcube_resampling.constants import AffineTransformMatrix, FloatInt

from .assertions import assert_given, assert_instance, assert_true
from .undefined import UNDEFINED


def _to_int_or_float(x: FloatInt) -> FloatInt:
    """If x is an int or is close to an int return it
    as int otherwise as float. Helps o avoid errors
    introduced by inaccurate floating point ops.
    """
    if isinstance(x, int):
        return x
    xf = float(x)
    xi = round(xf)
    return xi if math.isclose(xi, xf, rel_tol=1e-5) else xf


def _round_sequence(bbox: Sequence[FloatInt], ndigits: int = 7) -> Sequence[FloatInt]:
    return tuple(round(v, ndigits=ndigits) for v in bbox)


def _from_affine(matrix: affine.Affine) -> AffineTransformMatrix:
    return (matrix.a, matrix.b, matrix.c), (matrix.d, matrix.e, matrix.f)


def _to_affine(matrix: AffineTransformMatrix) -> affine.Affine:
    return affine.Affine(*matrix[0], *matrix[1])


def _normalize_crs(crs: str | pyproj.CRS) -> pyproj.CRS:
    if isinstance(crs, pyproj.CRS):
        return crs
    assert_instance(crs, str, "crs")
    return pyproj.CRS.from_string(crs)


def _normalize_int_pair(
    value: Any, name: str = None, default: tuple[int, int] | None = UNDEFINED
) -> tuple[int, int]:
    if isinstance(value, int):
        return value, value
    elif value is not None:
        x, y = value
        return int(x), int(y)
    elif default != UNDEFINED:
        return default
    else:
        assert_given(name, "name")
        raise ValueError(f"{name} must be an int or a sequence of two ints")


def _normalize_number_pair(
    value: Any, name: str = None, default: tuple[FloatInt, FloatInt] | None = UNDEFINED
) -> tuple[FloatInt, FloatInt]:
    if isinstance(value, (float, int)):
        x, y = value, value
        return _to_int_or_float(x), _to_int_or_float(y)
    elif value is not None:
        x, y = value
        return _to_int_or_float(x), _to_int_or_float(y)
    elif default != UNDEFINED:
        return default
    else:
        assert_given(name, "name")
        raise ValueError(f"{name} must be a number or a sequence of two numbers")


def to_lon_360(lon_var: np.ndarray | da.Array | xr.DataArray):
    if isinstance(lon_var, xr.DataArray):
        return lon_var.where(lon_var >= 0.0, lon_var + 360.0)
    else:
        lon_var = da.asarray(lon_var)
        return da.where(lon_var >= 0.0, lon_var, lon_var + 360.0)


def from_lon_360(lon_var: np.ndarray | da.Array | xr.DataArray):
    if isinstance(lon_var, xr.DataArray):
        return lon_var.where(lon_var <= 180.0, lon_var - 360.0)
    else:
        lon_var = da.asarray(lon_var)
        return da.where(lon_var <= 180.0, lon_var, lon_var - 360.0)


def get_dataset_chunks(dataset: xr.Dataset) -> dict[Hashable, int]:
    """Get the most common chunk sizes for each
    chunked dimension of *dataset*.

    Note: Only data variables are considered.

    Args:
        dataset: A dataset.

    Returns:
        A dictionary that maps dimension names to common chunk sizes.
    """

    # Record the frequencies of chunk sizes for
    # each dimension d in each data variable var
    dim_size_counts: dict[Hashable, dict[int, int]] = {}
    for var_name, var in dataset.data_vars.items():
        if var.chunks:
            for d, c in zip(var.dims, var.chunks):
                # compute max chunk size max_c from
                # e.g.  c = (512, 512, 512, 193)
                max_c = max(0, *c)
                # for dimension d, save the frequencies
                # of the different max_c
                if d not in dim_size_counts:
                    size_counts = {max_c: 1}
                    dim_size_counts[d] = size_counts
                else:
                    size_counts = dim_size_counts[d]
                    if max_c not in size_counts:
                        size_counts[max_c] = 1
                    else:
                        size_counts[max_c] += 1

    # For each dimension d, determine the most frequently
    # seen chunk size max_c
    dim_sizes: dict[Hashable, int] = {}
    for d, size_counts in dim_size_counts.items():
        max_count = 0
        best_max_c = 0
        for max_c, count in size_counts.items():
            if count > max_count:
                # Should always come here, because count=1 is minimum
                max_count = count
                best_max_c = max_c
        assert best_max_c > 0
        dim_sizes[d] = best_max_c

    return dim_sizes


def _default_xy_var_names(crs: pyproj.crs.CRS) -> tuple[str, str]:
    return ("lon", "lat") if crs.is_geographic else ("x", "y")


def _default_xy_dim_names(crs: pyproj.crs.CRS) -> tuple[str, str]:
    return _default_xy_var_names(crs)


def _assert_valid_xy_names(value: Any, name: str = None):
    assert_instance(value, tuple, name=name)
    assert_true(
        len(value) == 2 and all(value) and value[0] != value[1],
        f"invalid {name or 'value'}",
    )


def _assert_valid_xy_coords(xy_coords: Any):
    assert_instance(xy_coords, xr.DataArray, name="xy_coords")
    assert_true(
        xy_coords.ndim == 3
        and xy_coords.shape[0] == 2
        and xy_coords.shape[1] >= 2
        and xy_coords.shape[2] >= 2,
        "xy_coords must have dimensions"
        " (2, height, width) with height >= 2 and width >= 2",
    )


_RESOLUTIONS = {
    10: (1, 0),
    20: (2, 0),
    25: (25, 1),
    50: (5, 0),
    100: (1, -1),
}

_RESOLUTION_SET = {k / 100 for k in _RESOLUTIONS.keys()}


def round_to_fraction(value: float, digits: int = 2, resolution: float = 1) -> Fraction:
    """Round *value* at position given by significant
    *digits* and return result as fraction.

    Args:
        value: The value
        digits: The number of significant digits. Must be an integer >=
            1. Default is 2.
        resolution: The rounding resolution for the least significant
            digit. Must be one of (0.1, 0.2, 0.25, 0.5, 1). Default is
            1.

    Returns:
        The rounded value as fraction.Fraction instance.
    """
    if digits < 1:
        raise ValueError("digits must be a positive integer")
    resolution_key = round(100 * resolution)
    if resolution_key not in _RESOLUTIONS or not math.isclose(
        100 * resolution, resolution_key
    ):
        raise ValueError(f"resolution must be one of {_RESOLUTION_SET}")
    if value == 0:
        return Fraction(0, 1)
    sign = 1
    if value < 0:
        sign = -1
        value = -value
    resolution, resolution_digits = _RESOLUTIONS[resolution_key]
    exponent = math.floor(math.log10(value)) - digits - resolution_digits
    if exponent >= 0:
        magnitude = Fraction(10**exponent, 1)
    else:
        magnitude = Fraction(1, 10**-exponent)
    scaled_value = value / magnitude
    discrete_value = resolution * round(scaled_value / resolution)
    return (sign * discrete_value) * magnitude


def scale_xy_res_and_size(
    xy_res: tuple[float, float], size: tuple[int, int], xy_scale: tuple[float, float]
) -> tuple[tuple[float, float], tuple[int, int]]:
    """Scale given *xy_res* and *size* using *xy_scale*.
    Make sure, size components are not less than 2.
    """
    x_res, y_res = xy_res
    x_scale, y_scale = xy_scale
    w, h = size
    w, h = round(x_scale * w), round(y_scale * h)
    return (
        (x_res / x_scale, y_res / y_scale),
        (w if w >= 2 else 2, h if h >= 2 else 2),
    )



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""This module comprises reducer functions to be passed to
`dask.array.coarsen()` that either do not exist in numpy, like `mode`,
or have special implementations compared to their numpy equivalents.
"""

import warnings

import numba as nb
import numpy as np

_ALL = slice(None)

_DOC = """Computes the {property} of the windows in `block`.

Args:
    block: Array block from `dask.array.coarsen()` reshaped into smaller 
        windows to be reduced to size one. For spatial images, its shape will 
        be `(reduced_height, window_size_y, reduced_width, window_size_x)`.
    axis: A tuple providing the indexes of the window dimensions in the shape
        of `block`. For spatial images, this will be `(1, 3)`.

Returns:
    The reduced array containing the {property} of the windows 
    from `block`. For spatial images, its shape will be 
    `(reduced_height, reduced_width)`.
"""


def first(block: np.ndarray, axis: tuple[int, ...] | None = None) -> np.ndarray:
    if axis is None:
        return block  # edge block, pass through
    index = tuple(0 if i in axis else _ALL for i in range(block.ndim))
    return block[index]


def last(block: np.ndarray, axis: tuple[int, ...] | None = None) -> np.ndarray:
    if axis is None:
        return block  # edge block, pass through
    index = tuple(-1 if i in axis else _ALL for i in range(block.ndim))
    return block[index]


def center(block: np.ndarray, axis: tuple[int, ...] | None = None) -> np.ndarray:
    if axis is None:
        return block  # edge block, pass through
    shape = block.shape
    index = tuple(shape[i] // 2 if i in axis else _ALL for i in range(block.ndim))
    return block[index]


def mean(block: np.ndarray, axis: tuple[int, ...] | None = None) -> np.ndarray:
    return _reduce(np.mean, np.nanmean, block, axis)


def median(block: np.ndarray, axis: tuple[int, ...] | None = None) -> np.ndarray:
    return _reduce(np.median, np.nanmedian, block, axis)


def std(block: np.ndarray, axis: tuple[int, ...] | None = None) -> np.ndarray:
    return _reduce(np.std, np.nanstd, block, axis)


# noinspection PyShadowingBuiltins
def sum(block: np.ndarray, axis: tuple[int, ...] | None = None) -> np.ndarray:
    return _reduce(np.sum, np.nansum, block, axis)


def var(block: np.ndarray, axis: tuple[int, ...] | None = None) -> np.ndarray:
    return _reduce(np.var, np.nanvar, block, axis)


def _reduce(
    reducer, nan_reducer, block: np.ndarray, axis: tuple[int, ...] | None = None
) -> np.ndarray:
    if axis is None:
        # edge block, pass through
        return block
    elif np.issubdtype(block.dtype, np.floating):
        # For floating point types use nan-reducer
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            return nan_reducer(block, axis)
    else:
        # For integer and boolean types use "normal" reducer
        a = reducer(block, axis)
        if np.issubdtype(a.dtype, np.floating):
            # If result is floating point,
            # round and cast to original type
            return np.rint(a).astype(block.dtype)
        return a


def mode(block: np.ndarray, axis: tuple[int, ...] | None = None) -> np.ndarray:
    if axis is None:
        return block  # edge block, pass through

    # This implementation assumes that `block` contains categorical
    # numbers as it computes most frequent values. It does therefore
    # neither use fuzzy equality comparisons nor checks for NaN should
    # floating point data be passed.

    ndim = len(axis)  # number of dimensions in which to reduce
    block = np.moveaxis(block, axis, range(-ndim, 0))
    flat = block.reshape(-1, np.prod(block.shape[-ndim:]))

    min_val = int(flat.min())
    max_val = int(flat.max())
    mode_range = max_val - min_val + 1

    normalized = (flat - min_val).astype(np.int64)
    mode_indices = _mode_from_normalized(
        normalized, offset=min_val, mode_range=mode_range
    )
    return mode_indices.reshape(block.shape[:-ndim])


@nb.njit()
def _mode_from_normalized(
    flat_block: np.ndarray, offset: int, mode_range: int
) -> np.ndarray:  # pragma: no cover
    size = flat_block.shape[0]
    out = np.empty(size, dtype=np.int64)
    for i in range(size):
        counts = np.zeros(mode_range, dtype=np.int64)
        for val in flat_block[i]:
            counts[val] += 1
        mode_val = 0
        max_count = counts[0]
        for j in range(1, mode_range):
            if counts[j] > max_count:
                max_count = counts[j]
                mode_val = j
        out[i] = mode_val + offset
    return out


first.__doc__ = _DOC.format(property="first value")
last.__doc__ = _DOC.format(property="last value")
center.__doc__ = _DOC.format(property="center value")
mean.__doc__ = _DOC.format(property="mean")
median.__doc__ = _DOC.format(property="median")
mode.__doc__ = _DOC.format(property="mode (most frequent value)")
std.__doc__ = _DOC.format(property="standard deviation")
sum.__doc__ = _DOC.format(property="sum")
var.__doc__ = _DOC.format(property="variance")



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest
from fractions import Fraction

import dask.array as da
import numpy as np
import xarray as xr

# noinspection PyProtectedMember
from xcube_resampling.gridmapping.helpers import (
    _normalize_number_pair,
    _to_int_or_float,
    from_lon_360,
    round_to_fraction,
    to_lon_360,
)


class RoundToFractionTest(unittest.TestCase):
    dump = False

    def test_invalid(self):
        with self.assertRaises(ValueError):
            round_to_fraction(0.29, digits=0)
        with self.assertRaises(ValueError):
            round_to_fraction(0.29, resolution=0)
        with self.assertRaises(ValueError):
            round_to_fraction(0.29, resolution=0.12)

    def test_1_025(self):
        def f(value):
            return float(round_to_fraction(value, 1, 0.25))

        self.assertAlmostEqual(-1.0, f(-1))
        self.assertAlmostEqual(0.0, f(0))
        self.assertAlmostEqual(1.0, f(1))
        self.assertAlmostEqual(1.25, f(1.2))
        self.assertAlmostEqual(1.25, f(1.3))
        self.assertAlmostEqual(1.5, f(1.4))
        self.assertAlmostEqual(1.5, f(1.45))
        self.assertAlmostEqual(1.5, f(1.51))
        self.assertAlmostEqual(1.75, f(1.7))
        self.assertAlmostEqual(2.0, f(1.9))
        self.assertAlmostEqual(2.0, f(1.96))
        self.assertAlmostEqual(2.0, f(1.98))
        self.assertAlmostEqual(2.0, f(2))

    def test_2_025(self):
        def f(value):
            return float(round_to_fraction(value, 2, 0.25))

        self.assertAlmostEqual(-1.0, f(-1))
        self.assertAlmostEqual(0.0, f(0))
        self.assertAlmostEqual(1.0, f(1))
        self.assertAlmostEqual(1.2, f(1.2))
        self.assertAlmostEqual(1.225, f(1.23))
        self.assertAlmostEqual(1.3, f(1.3))
        self.assertAlmostEqual(1.4, f(1.4))
        self.assertAlmostEqual(1.45, f(1.45))
        self.assertAlmostEqual(1.5, f(1.51))
        self.assertAlmostEqual(1.7, f(1.7))
        self.assertAlmostEqual(1.8, f(1.79))
        self.assertAlmostEqual(1.9, f(1.9))
        self.assertAlmostEqual(1.95, f(1.96))
        self.assertAlmostEqual(1.975, f(1.98))
        self.assertAlmostEqual(2.0, f(2))

    def test_default(self):
        values = [
            [-1.0, -1.0, Fraction(-1, 1)],
            [0.0, 0.0, Fraction(0, 1)],
            [5.247476065426347e-09, 5.2e-09, Fraction(13, 2500000000)],
            [3.427467229408875e-06, 3.4e-06, Fraction(17, 5000000)],
            [4.501758583626108e-06, 4.5e-06, Fraction(9, 2000000)],
            [1.1351705264714663e-05, 1.1e-05, Fraction(11, 1000000)],
            [0.00048171747406886744, 0.00048, Fraction(3, 6250)],
            [0.0018032657496927416, 0.0018, Fraction(9, 5000)],
            [0.0019897341919324425, 0.002, Fraction(1, 500)],
            [0.0041643509375105065, 0.0042, Fraction(21, 5000)],
            [0.030607346091352187, 0.031, Fraction(31, 1000)],
            [1.0076973439575128, 1.0, Fraction(1, 1)],
            [1.0, 1.0, Fraction(1, 1)],
            [84.54360269093455, 85.0, Fraction(85, 1)],
            [494.86581234602096, 490.0, Fraction(490, 1)],
            [987.9441243998718, 990.0, Fraction(990, 1)],
            [1757.368043916636, 1800.0, Fraction(1800, 1)],
            [1143506.2928512183, 1100000.0, Fraction(1100000, 1)],
            [217971970.75235566, 220000000.0, Fraction(220000000, 1)],
        ]
        self._assert_values(values, dict())

    def test_3_025(self):
        actual = round_to_fraction(1, digits=1, resolution=0.25)
        self.assertEqual(Fraction(1, 1), actual)
        values = [
            [-1.0, -1.0, Fraction(-1, 1)],
            [0.0, 0.0, Fraction(0, 1)],
            [5.247476065426347e-09, 5.2475e-09, Fraction(2099, 400000000000)],
            [3.427467229408875e-06, 3.4275e-06, Fraction(1371, 400000000)],
            [4.501758583626108e-06, 4.5025e-06, Fraction(1801, 400000000)],
            [1.1351705264714663e-05, 1.135e-05, Fraction(227, 20000000)],
            [0.00048171747406886744, 0.00048175, Fraction(1927, 4000000)],
            [0.0018032657496927416, 0.0018025, Fraction(721, 400000)],
            [0.0019897341919324425, 0.00199, Fraction(199, 100000)],
            [0.0041643509375105065, 0.004165, Fraction(833, 200000)],
            [0.030607346091352187, 0.0306, Fraction(153, 5000)],
            [1.0076973439575128, 1.0075, Fraction(403, 400)],
            [1.0, 1.0, Fraction(1, 1)],
            [84.54360269093455, 84.55, Fraction(1691, 20)],
            [494.86581234602096, 494.75, Fraction(1979, 4)],
            [987.9441243998718, 988.0, Fraction(988, 1)],
            [1757.368043916636, 1757.5, Fraction(3515, 2)],
            [1143506.2928512183, 1142500.0, Fraction(1142500, 1)],
            [217971970.75235566, 218000000.0, Fraction(218000000, 1)],
        ]
        self._assert_values(values, dict(digits=3, resolution=0.25))

    def test_2_5(self):
        values = [
            [-1.0, -1.0, Fraction(-1, 1)],
            [0.0, 0.0, Fraction(0, 1)],
            [5.247476065426347e-09, 5.25e-09, Fraction(21, 4000000000)],
            [3.427467229408875e-06, 3.45e-06, Fraction(69, 20000000)],
            [4.501758583626108e-06, 4.5e-06, Fraction(9, 2000000)],
            [1.1351705264714663e-05, 1.15e-05, Fraction(23, 2000000)],
            [0.00048171747406886744, 0.00048, Fraction(3, 6250)],
            [0.0018032657496927416, 0.0018, Fraction(9, 5000)],
            [0.0019897341919324425, 0.002, Fraction(1, 500)],
            [0.0041643509375105065, 0.00415, Fraction(83, 20000)],
            [0.030607346091352187, 0.0305, Fraction(61, 2000)],
            [1.0076973439575128, 1.0, Fraction(1, 1)],
            [1.0, 1.0, Fraction(1, 1)],
            [84.54360269093455, 84.5, Fraction(169, 2)],
            [494.86581234602096, 495.0, Fraction(495, 1)],
            [987.9441243998718, 990.0, Fraction(990, 1)],
            [1757.368043916636, 1750.0, Fraction(1750, 1)],
            [1143506.2928512183, 1150000.0, Fraction(1150000, 1)],
            [217971970.75235566, 220000000.0, Fraction(220000000, 1)],
        ]
        self._assert_values(values, dict(digits=2, resolution=0.5))

    def _assert_values(self, values, kwargs):
        if self.dump:
            results = []
            for value, _, _ in values:
                actual_fraction = round_to_fraction(value, **kwargs)
                results.append([value, float(actual_fraction), actual_fraction])
            import pprint

            pprint.pprint(results)

        for value, expected_float, expected_fraction in values:
            actual_fraction = round_to_fraction(value, **kwargs)
            self.assertEqual(expected_fraction, actual_fraction)
            self.assertAlmostEqual(expected_float, float(actual_fraction))


class ToIntOrFloatTest(unittest.TestCase):
    def test_down_to_int(self):
        result = _to_int_or_float(90.0001)
        self.assertEqual(90, result)

    def test_leave_as_bigger_float(self):
        result = _to_int_or_float(90.001)
        self.assertEqual(90.001, result)

    def test_up_to_int(self):
        result = _to_int_or_float(89.9999)
        self.assertEqual(90, result)

    def test_leave_as_smaller_float(self):
        result = _to_int_or_float(89.999)
        self.assertEqual(89.999, result)

    def test_up_to_int_small_value(self):
        result = _to_int_or_float(0.99999)
        self.assertEqual(1, result)

    def test_leave_as_smaller_float_small_value(self):
        result = _to_int_or_float(0.9999)
        self.assertEqual(0.9999, result)


class TestNormalizeNumberPair(unittest.TestCase):

    def test_single_number(self):
        result = _normalize_number_pair(5)
        self.assertEqual(result, (5, 5))
        result = _normalize_number_pair(3.5)
        self.assertEqual(result, (3.5, 3.5))

    def test_pair_of_numbers(self):
        result = _normalize_number_pair((2, 4))
        self.assertEqual(result, (2, 4))
        result = _normalize_number_pair((1.5, 2.5))
        self.assertEqual(result, (1.5, 2.5))

    def test_default_value(self):
        default_pair = (10, 20)
        result = _normalize_number_pair(None, default=default_pair)
        self.assertEqual(result, default_pair)

    def test_value_error_when_no_value_or_default(self):
        with self.assertRaises(ValueError) as cm:
            _normalize_number_pair(None, name="test_var")
        self.assertIn(
            "test_var must be a number or a sequence of two numbers", str(cm.exception)
        )


class TestToLon360(unittest.TestCase):

    def test_numpy_array(self):
        arr = np.array([-10, 0, 45, 190, -180])
        expected = np.array([350, 0, 45, 190, 180])
        result = (
            to_lon_360(arr).compute()
            if isinstance(to_lon_360(arr), da.Array)
            else to_lon_360(arr)
        )
        np.testing.assert_array_equal(result, expected)

    def test_dask_array(self):
        arr = da.from_array(np.array([-10, 0, 45, 190, -180]), chunks=2)
        expected = np.array([350, 0, 45, 190, 180])
        result = to_lon_360(arr).compute()
        np.testing.assert_array_equal(result, expected)

    def test_xarray_dataarray(self):
        arr = xr.DataArray([-10, 0, 45, 190, -180])
        expected = np.array([350, 0, 45, 190, 180])
        result = to_lon_360(arr).values
        np.testing.assert_array_equal(result, expected)


class TestFromLon360(unittest.TestCase):

    def test_numpy_array(self):
        arr = np.array([350, 0, 45, 190, 180])
        expected = np.array([-10, 0, 45, -170, 180])
        result = (
            from_lon_360(arr).compute()
            if isinstance(from_lon_360(arr), da.Array)
            else from_lon_360(arr)
        )
        np.testing.assert_array_equal(result, expected)

    def test_dask_array(self):
        arr = da.from_array(np.array([350, 0, 45, 190, 180]), chunks=2)
        expected = np.array([-10, 0, 45, -170, 180])
        result = from_lon_360(arr).compute()
        np.testing.assert_array_equal(result, expected)

    def test_xarray_dataarray(self):
        arr = xr.DataArray([350, 0, 45, 190, 180])
        expected = np.array([-10, 0, 45, -170, 180])
        result = from_lon_360(arr).values
        np.testing.assert_array_equal(result, expected)



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import abc

import dask
import dask.array as da
import numpy as np
import pyproj
import xarray as xr

from xcube_resampling.constants import FloatInt

from .assertions import assert_instance, assert_true
from .base import DEFAULT_TOLERANCE, GridMapping
from .helpers import (
    _assert_valid_xy_names,
    _default_xy_var_names,
    _normalize_crs,
    _normalize_int_pair,
    _round_sequence,
    _to_int_or_float,
    from_lon_360,
    round_to_fraction,
    to_lon_360,
)

_ER = 6371000


class CoordsGridMapping(GridMapping, abc.ABC):
    """
    Grid mapping constructed from 1D/2D coordinate
    variables and a CRS.
    """

    @property
    def x_coords(self):
        assert isinstance(self._x_coords, xr.DataArray)
        return self._x_coords

    @property
    def y_coords(self):
        assert isinstance(self._y_coords, xr.DataArray)
        return self._y_coords

    def _new_x_coords(self) -> xr.DataArray:
        # Should never come here
        return self._x_coords

    def _new_y_coords(self) -> xr.DataArray:
        # Should never come here
        return self._y_coords


class Coords1DGridMapping(CoordsGridMapping):
    """Grid mapping constructed from
    1D coordinate variables and a CRS.
    """

    def _new_xy_coords(self) -> xr.DataArray:
        y, x = xr.broadcast(self._y_coords, self._x_coords)
        tmp = xr.concat([x, y], dim="coord")
        return tmp.chunk(
            {dim: size for (dim, size) in zip(tmp.dims, self.xy_coords_chunks)}
        )


class Coords2DGridMapping(CoordsGridMapping):
    """Grid mapping constructed from
    2D coordinate variables and a CRS.
    """

    def _new_xy_coords(self) -> xr.DataArray:
        tmp = xr.concat([self._x_coords, self._y_coords], dim="coord")
        return tmp.chunk(
            {dim: size for (dim, size) in zip(tmp.dims, self.xy_coords_chunks)}
        )


def new_grid_mapping_from_coords(
    x_coords: xr.DataArray,
    y_coords: xr.DataArray,
    crs: str | pyproj.crs.CRS,
    *,
    xy_bbox: tuple[FloatInt, FloatInt, FloatInt, FloatInt] = None,
    tile_size: int | tuple[int, int] = None,
    tolerance: float = DEFAULT_TOLERANCE,
) -> GridMapping:
    crs = _normalize_crs(crs)
    assert_instance(x_coords, xr.DataArray, name="x_coords")
    assert_instance(y_coords, xr.DataArray, name="y_coords")
    assert_true(
        x_coords.ndim in (1, 2), "x_coords and y_coords must be either 1D or 2D arrays"
    )
    assert_instance(tolerance, float, name="tolerance")
    assert_true(tolerance > 0.0, "tolerance must be greater zero")

    if x_coords.name and y_coords.name:
        xy_var_names = str(x_coords.name), str(y_coords.name)
    else:
        xy_var_names = _default_xy_var_names(crs)

    tile_size = _normalize_int_pair(tile_size, default=None)

    if x_coords.ndim == 1 and y_coords.ndim == 1:
        gm = new_1d_grid_mapping_from_coords(
            x_coords,
            y_coords,
            crs,
            xy_var_names,
            xy_bbox=xy_bbox,
            tile_size=tile_size,
            tolerance=tolerance,
        )
    else:
        gm = new_2d_grid_mapping_from_coords(
            x_coords,
            y_coords,
            crs,
            xy_var_names,
            xy_bbox=xy_bbox,
            tile_size=tile_size,
            tolerance=tolerance,
        )
    return gm


def new_1d_grid_mapping_from_coords(
    x_coords: xr.DataArray,
    y_coords: xr.DataArray,
    crs: str | pyproj.crs.CRS,
    xy_var_names: tuple[str, str],
    *,
    xy_bbox: tuple[FloatInt, FloatInt, FloatInt, FloatInt] = None,
    tile_size: int | tuple[int, int] = None,
    tolerance: float = DEFAULT_TOLERANCE,
) -> Coords1DGridMapping:

    assert_true(
        x_coords.size >= 2 and y_coords.size >= 2,
        "sizes of x_coords and y_coords 1D arrays must be >= 2",
    )

    is_lon_360 = None
    if crs.is_geographic:
        is_lon_360 = bool(da.any(x_coords > 180))
    y_diff = np.diff(y_coords)
    x_diff = np.diff(x_coords)
    if np.any(np.nanmax(abs(x_diff)) > 180) and not is_lon_360 and crs.is_geographic:
        x_coords = to_lon_360(x_coords)
        x_diff = np.diff(x_coords)
        is_lon_360 = True

    x_res = x_diff[0]
    y_res = y_diff[0]
    is_regular = (
        da.allclose(x_diff, x_res, atol=tolerance)
        and da.allclose(y_diff, y_res, atol=tolerance)
    ).compute()
    if is_regular:
        x_res = round_to_fraction(float(x_res), 5, 0.1)
        y_res = round_to_fraction(abs(float(y_res)), 5, 0.1)
    else:
        x_res = round_to_fraction(float(np.nanmedian(x_diff, axis=0)), 5, 0.1)
        y_res = round_to_fraction(abs(float(np.nanmedian(y_diff, axis=0))), 5, 0.1)

    if (
        tile_size is None
        and x_coords.chunks is not None
        and y_coords.chunks is not None
    ):
        tile_size = (max(0, *x_coords.chunks[0]), max(0, *y_coords.chunks[0]))

    # Guess j axis direction
    is_j_axis_up = bool(y_coords[0] < y_coords[-1])

    if xy_bbox is None:
        x_vals, y_vals = dask.compute(x_coords[[0, -1]], y_coords[[0, -1]])
        x_min, x_max = x_vals.values
        y_min, y_max = y_vals.values
        if not is_j_axis_up:
            y_min, y_max = y_max, y_min
        x_pad = x_res / 2
        y_pad = y_res / 2
        x_min = _to_int_or_float(x_min - x_pad)
        y_min = _to_int_or_float(y_min - y_pad)
        x_max = _to_int_or_float(x_max + x_pad)
        y_max = _to_int_or_float(y_max + y_pad)
        xy_bbox = _round_sequence((x_min, y_min, x_max, y_max))

    return Coords1DGridMapping(
        x_coords=x_coords,
        y_coords=y_coords,
        crs=crs,
        size=(x_coords.size, y_coords.size),
        tile_size=tile_size,
        xy_bbox=xy_bbox,
        xy_res=(x_res, y_res),
        xy_var_names=xy_var_names,
        xy_dim_names=(str(x_coords.dims[0]), str(y_coords.dims[0])),
        is_regular=is_regular,
        is_lon_360=is_lon_360,
        is_j_axis_up=is_j_axis_up,
    )


def new_2d_grid_mapping_from_coords(
    x_coords: xr.DataArray,
    y_coords: xr.DataArray,
    crs: str | pyproj.crs.CRS,
    xy_var_names: tuple[str, str],
    *,
    xy_bbox: tuple[FloatInt, FloatInt, FloatInt, FloatInt] = None,
    tile_size: int | tuple[int, int] = None,
    tolerance: float = DEFAULT_TOLERANCE,
) -> Coords2DGridMapping:

    assert_true(
        x_coords.shape == y_coords.shape,
        "shapes of x_coords and y_coords 2D arrays must be equal",
    )
    assert_true(
        x_coords.dims == y_coords.dims,
        "dimensions of x_coords and y_coords 2D arrays must be equal",
    )

    height, width = x_coords.shape
    ydim, xdim = x_coords.dims
    x_da, y_da = dask.compute(
        x_coords.isel({ydim: [0, -1], xdim: [0, 1, -1]}),
        y_coords.isel({ydim: [0, 1, -1], xdim: [0, -1]}),
    )
    x_vals = x_da.values
    y_vals = y_da.values
    x00 = x_vals[0, 0]
    x00_diff = x_vals[0, 1] - x_vals[0, 0]
    x01 = x_vals[0, 2]
    x10 = x_vals[1, 0]
    x11 = x_vals[1, 2]
    y00 = y_vals[0, 0]
    y01 = y_vals[0, 1]
    y00_diff = abs(y_vals[0, 0] - y_vals[1, 0])
    y10 = y_vals[2, 0]
    y11 = y_vals[2, 1]

    is_j_axis_up = y00 < y10
    x_max = max(x00, x01, x10, x11)
    x_min = min(x00, x01, x10, x11)
    x_span = x_max - x_min
    is_lon_360 = False
    if crs.is_geographic:
        is_lon_360 = bool(x_max > 180)
    if x_span > 180:
        x_coords = to_lon_360(x_coords)
        x_da = x_coords.isel({ydim: [0, 1, -1], xdim: [0, -1]}).compute()
        x_vals = x_da.values
        x00 = x_vals[0, 0]
        x00_diff = x_vals[1, 0] - x_vals[0, 0]
        x01 = x_vals[0, 1]
        x10 = x_vals[2, 0]
        x11 = x_vals[2, 1]
        is_lon_360 = True

    x_res = abs(np.mean([x00, x10]) - np.mean([x11, x01])) / (width - 1)
    y_res = abs(np.mean([y00, y01]) - np.mean([y11, y10])) / (height - 1)
    assert_true(
        x_res > 0 and y_res > 0,
        "internal error: x_res and y_res could not be determined",
        exception_type=RuntimeError,
    )
    is_regular = (
        np.isclose(x00_diff, x_res, atol=tolerance)
        and np.isclose(x00, x10, atol=tolerance)
        and np.isclose(x01, x11, atol=tolerance)
        and np.isclose(y00_diff, y_res, atol=tolerance)
        and np.isclose(y00, y01, atol=tolerance)
        and np.isclose(y10, y11, atol=tolerance)
    )
    x_res = round_to_fraction(x_res, 5, 0.1)
    y_res = round_to_fraction(y_res, 5, 0.1)

    if tile_size is None and x_coords.chunks is not None:
        j_chunks, i_chunks = x_coords.chunks
        tile_size = max(0, *i_chunks), max(0, *j_chunks)

    if tile_size is not None:
        tile_width, tile_height = tile_size
        x_coords = x_coords.chunk(
            {x_coords.dims[0]: tile_height, x_coords.dims[1]: tile_width}
        )
        y_coords = y_coords.chunk(
            {y_coords.dims[0]: tile_height, y_coords.dims[1]: tile_width}
        )

    if xy_bbox is None:
        x_pad, y_pad = x_res / 2, y_res / 2
        x_min = min(x00, x10, x01, x11)
        x_max = max(x00, x10, x01, x11)
        y_min = min(y00, y01, y10, y11)
        y_max = max(y00, y01, y10, y11)
        x_min = _to_int_or_float(x_min - x_pad)
        y_min = _to_int_or_float(y_min - y_pad)
        x_max = _to_int_or_float(x_max + x_pad)
        y_max = _to_int_or_float(y_max + y_pad)
        xy_bbox = _round_sequence((x_min, y_min, x_max, y_max))

    return Coords2DGridMapping(
        x_coords=x_coords,
        y_coords=y_coords,
        crs=crs,
        size=(width, height),
        tile_size=tile_size,
        xy_bbox=xy_bbox,
        xy_res=(x_res, y_res),
        xy_var_names=xy_var_names,
        xy_dim_names=(str(x_coords.dims[1]), str(x_coords.dims[0])),
        is_regular=is_regular,
        is_lon_360=is_lon_360,
        is_j_axis_up=is_j_axis_up,
    )


def grid_mapping_to_coords(
    grid_mapping: GridMapping,
    xy_var_names: tuple[str, str] = None,
    xy_dim_names: tuple[str, str] = None,
    reuse_coords: bool = False,
    exclude_bounds: bool = False,
) -> dict[str, xr.DataArray]:
    """Get CF-compliant axis coordinate variables and cell
    boundary coordinate variables.

    Defined only for grid mappings with regular x,y coordinates.

    Args:
        grid_mapping: A regular grid mapping.
        xy_var_names: Optional coordinate variable names (x_var_name,
            y_var_name).
        xy_dim_names: Optional coordinate dimensions names (x_dim_name,
            y_dim_name).
        reuse_coords: Whether to either reuse target coordinate arrays
            from target_gm or to compute new ones.
        exclude_bounds: If True, do not create bounds coordinates.
            Defaults to False. Ignored if *reuse_coords* is True.

    Returns:
        dictionary with coordinate variables
    """

    if xy_var_names:
        _assert_valid_xy_names(xy_var_names, name="xy_var_names")
    if xy_dim_names:
        _assert_valid_xy_names(xy_dim_names, name="xy_dim_names")

    if reuse_coords:
        try:
            # noinspection PyUnresolvedReferences
            x, y = grid_mapping.x_coords, grid_mapping.y_coords
        except AttributeError:
            x, y = None, None
        if (
            isinstance(x, xr.DataArray)
            and isinstance(y, xr.DataArray)
            and x.ndim == 1
            and y.ndim == 1
            and x.size == grid_mapping.width
            and y.size == grid_mapping.height
        ):
            return {
                name: xr.DataArray(coord.values, dims=dim, attrs=coord.attrs)
                for name, dim, coord in zip(xy_var_names, xy_dim_names, (x, y))
            }

    x_name, y_name = xy_var_names or grid_mapping.xy_var_names
    x_dim_name, y_dim_name = xy_dim_names or grid_mapping.xy_dim_names
    w, h = grid_mapping.size
    x1, y1, x2, y2 = grid_mapping.xy_bbox
    x_res, y_res = grid_mapping.xy_res
    x_res_05 = x_res / 2
    y_res_05 = y_res / 2

    dtype = np.float64

    x_data = np.linspace(x1 + x_res_05, x2 - x_res_05, w, dtype=dtype)
    if grid_mapping.is_lon_360:
        x_data = from_lon_360(x_data)

    if grid_mapping.is_j_axis_up:
        y_data = np.linspace(y1 + y_res_05, y2 - y_res_05, h, dtype=dtype)
    else:
        y_data = np.linspace(y2 - y_res_05, y1 + y_res_05, h, dtype=dtype)

    if grid_mapping.crs.is_geographic:
        x_attrs = dict(
            long_name="longitude coordinate",
            standard_name="longitude",
            units="degrees_east",
        )
        y_attrs = dict(
            long_name="latitude coordinate",
            standard_name="latitude",
            units="degrees_north",
        )
    else:
        x_attrs = dict(
            long_name="x coordinate of projection",
            standard_name="projection_x_coordinate",
        )
        y_attrs = dict(
            long_name="y coordinate of projection",
            standard_name="projection_y_coordinate",
        )

    x_coords = xr.DataArray(x_data, dims=x_dim_name, attrs=x_attrs)
    y_coords = xr.DataArray(y_data, dims=y_dim_name, attrs=y_attrs)
    coords = {
        x_name: x_coords,
        y_name: y_coords,
    }
    if not exclude_bounds:
        x_bnds_0_data = np.linspace(x1, x2 - x_res, w, dtype=dtype)
        x_bnds_1_data = np.linspace(x1 + x_res, x2, w, dtype=dtype)

        if grid_mapping.is_lon_360:
            x_bnds_0_data = from_lon_360(x_bnds_0_data)
            x_bnds_1_data = from_lon_360(x_bnds_1_data)

        if grid_mapping.is_j_axis_up:
            y_bnds_0_data = np.linspace(y1, y2 - y_res, h, dtype=dtype)
            y_bnds_1_data = np.linspace(y1 + y_res, y2, h, dtype=dtype)
        else:
            y_bnds_0_data = np.linspace(y2, y1 + y_res, h, dtype=dtype)
            y_bnds_1_data = np.linspace(y2 - y_res, y1, h, dtype=dtype)

        bnds_dim_name = "bnds"
        x_bnds_name = f"{x_name}_{bnds_dim_name}"
        y_bnds_name = f"{y_name}_{bnds_dim_name}"
        # Note, according to CF, bounds variables are not required to have
        # any attributes, so we don't pass any.
        x_bnds_coords = xr.DataArray(
            list(zip(x_bnds_0_data, x_bnds_1_data)), dims=[x_dim_name, bnds_dim_name]
        )
        y_bnds_coords = xr.DataArray(
            list(zip(y_bnds_0_data, y_bnds_1_data)), dims=[y_dim_name, bnds_dim_name]
        )
        x_coords.attrs.update(bounds=x_bnds_name)
        y_coords.attrs.update(bounds=y_bnds_name)
        coords.update(
            {
                x_bnds_name: x_bnds_coords,
                y_bnds_name: y_bnds_coords,
            }
        )

    return coords



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import math
from collections.abc import Hashable, Iterable, Mapping

import dask.array as da
import numpy as np
import pyproj
import xarray as xr

from .affine import affine_transform_dataset
from .constants import (
    SCALE_LIMIT,
    FillValues,
    PreventNaNPropagations,
    SpatialAggMethods,
    SpatialInterpMethod,
    SpatialInterpMethods,
    SpatialInterpMethodStr,
)
from .gridmapping import GridMapping
from .gridmapping.helpers import from_lon_360
from .utils import (
    SourceTileIndexing,
    _clip_if_needed,
    _get_fill_value,
    _get_spatial_interp_method_str,
    _get_spatial_interp_methods,
    _prep_spatial_interp_methods_downscale,
    _reorganize_tiled_array,
    _sample_array_at_indices,
    _select_variables,
    normalize_grid_mapping,
)


def reproject_dataset(
    source_ds: xr.Dataset,
    target_gm: GridMapping,
    *,
    source_gm: GridMapping | None = None,
    variables: str | Iterable[str] | None = None,
    interp_methods: SpatialInterpMethods | None = None,
    agg_methods: SpatialAggMethods | None = None,
    prevent_nan_propagations: PreventNaNPropagations = False,
    fill_values: FillValues | None = None,
) -> xr.Dataset:
    """
    Reproject a dataset from one coordinate reference system (CRS) to another.

    This function transforms a dataset’s 2D spatial variables to match a new CRS and
    grid layout defined by `target_gm`. It handles interpolation, optional
    downsampling, and fill values for areas outside the input bounds.

    Args:
        source_ds: The input dataset to be reprojected.
        target_gm: The target grid mapping that defines the spatial reference and
            grid layout in the target CRS.
        source_gm: Optional source grid mapping of the input dataset. If not
            provided, it is inferred from the dataset.
        variables: Optional variable name or list of variable names to reproject.
            If None, all suitable variables are processed.
        interp_methods: Optional interpolation method to be used for upsampling spatial
            data variables. Can be a single interpolation method for all variables or a
            dictionary mapping variable names or dtypes to interpolation method.
            Supported methods include:

            - `0` (nearest neighbor)
            - `1` (linear / bilinear)
            - `"nearest"`
            - `"triangular"`
            - `"bilinear"`

            The default is `0` for integer arrays, else `1`.
        agg_methods: Optional aggregation methods for downsampling spatial variables.
            Can be a single method for all variables or a dictionary mapping variable
            names or dtypes to methods. Supported methods include:
                "center", "count", "first", "last", "max", "mean", "median",
                "mode", "min", "prod", "std", "sum", and "var".
            Defaults to "center" for integer arrays, else "mean".
        prevent_nan_propagations: Optional boolean or mapping to prevent NaN
            propagation during upsampling (only applies when interpolation method
            is not nearest). Can be a single boolean or a dictionary mapping
            variable names or dtypes to booleans. Defaults to False.
        fill_values: Optional fill value(s) to assign outside source coverage.
            Can be a single value or dictionary by variable or type. If not set,
            defaults are:

            - float: NaN
            - uint8: 255
            - uint16: 65535
            - other integers: -1

    Returns:
        A new dataset with variables reprojected to the target CRS and
            grid. Variables without 2D spatial dimensions are copied as-is.
            1D spatial coordinate variables are ignored in the output.
    """
    source_ds = _select_variables(source_ds, variables)

    source_gm = source_gm or GridMapping.from_dataset(source_ds)
    source_ds = normalize_grid_mapping(source_ds, source_gm)

    if source_gm.is_j_axis_up:
        v_var = source_gm.xy_var_names[1]
        source_ds = source_ds.isel({v_var: slice(None, None, -1)})
        source_gm = GridMapping.from_dataset(source_ds)

    transformer = pyproj.Transformer.from_crs(
        target_gm.crs, source_gm.crs, always_xy=True
    )
    source_ds, source_gm, is_empty = _clip_if_needed(
        source_ds, source_gm, target_gm, fill_values, transformer
    )
    if is_empty:
        return source_ds

    # If source has higher resolution than target, downscale first, then reproject
    interp_methods = _get_spatial_interp_methods(source_ds, interp_methods)
    source_ds, source_gm = _downscale_source_dataset(
        source_ds,
        source_gm,
        target_gm,
        transformer,
        interp_methods,
        agg_methods,
        prevent_nan_propagations,
    )

    # For each bounding box in the target grid mapping:
    # - determine the indices of the bbox in the source dataset
    # - compute the pad_width to handle areas requested by target_gm that exceed the
    #   bounds of source_gm.
    # - extract the corresponding coordinates for each bbox in the source dataset

    indexing, x_coords, y_coords = _get_src_bboxes_indices(
        transformer, source_gm, target_gm
    )

    # transform grid points from target grid mapping to source grid mapping
    source_gm_transformed = target_gm.transform(source_gm.crs)
    source_xx = source_gm_transformed.x_coords.data
    source_yy = source_gm_transformed.y_coords.data

    # reproject dataset
    target_ds = _assemble_reprojected_dataset(
        source_ds,
        source_gm,
        target_gm,
        source_xx,
        source_yy,
        x_coords,
        y_coords,
        indexing,
        interp_methods,
        fill_values,
    )

    return target_ds


def _assemble_reprojected_dataset(
    source_ds: xr.Dataset,
    source_gm: GridMapping,
    target_gm: GridMapping,
    source_xx: da.Array,
    source_yy: da.Array,
    x_coords: da.Array,
    y_coords: da.Array,
    indexing: SourceTileIndexing,
    interp_methods: SpatialInterpMethods | None = None,
    fill_values: FillValues | None = None,
):
    src_x_name, src_y_name = source_gm.xy_var_names
    tgt_x_name, tgt_y_name = target_gm.xy_var_names

    coords = source_ds.coords.to_dataset()
    coords = coords.drop_vars((src_x_name, src_y_name), errors="ignore")

    tgt_x = target_gm.x_coords
    if target_gm.is_lon_360:
        tgt_x = from_lon_360(tgt_x)
    coords[tgt_x_name] = tgt_x
    coords[tgt_y_name] = target_gm.y_coords
    coords["spatial_ref"] = xr.DataArray(0, attrs=target_gm.crs.to_cf())
    target_ds = xr.Dataset(coords=coords, attrs=source_ds.attrs)

    yx_dims = (source_gm.xy_dim_names[1], source_gm.xy_dim_names[0])
    for var_name, data_array in source_ds.items():
        if data_array.dims[-2:] == yx_dims:
            assert len(data_array.dims) in (
                2,
                3,
            ), f"Data variable {var_name} has {len(data_array.dims)} dimensions."

            target_ds[var_name] = _reproject_data_array(
                data_array,
                var_name,
                source_gm,
                target_gm,
                source_xx,
                source_yy,
                x_coords,
                y_coords,
                indexing,
                interp_methods,
                fill_values,
            )
        elif yx_dims[0] not in data_array.dims and yx_dims[1] not in data_array.dims:
            target_ds[var_name] = data_array

    return target_ds


def _downscale_source_dataset(
    source_ds: xr.Dataset,
    source_gm: GridMapping,
    target_gm: GridMapping,
    transformer: pyproj.Transformer,
    interp_methods: Mapping[Hashable, SpatialInterpMethod],
    agg_methods: SpatialAggMethods | None,
    prevent_nan_propagations: PreventNaNPropagations,
) -> tuple[xr.Dataset, GridMapping]:
    if all(v in (0, "nearest") for v in interp_methods.values()):
        return source_ds, source_gm

    bbox_trans = transformer.transform_bounds(*target_gm.xy_bbox)
    xres_trans = (bbox_trans[2] - bbox_trans[0]) / target_gm.width
    yres_trans = (bbox_trans[3] - bbox_trans[1]) / target_gm.height
    x_scale = source_gm.x_res / xres_trans
    y_scale = source_gm.y_res / yres_trans
    if x_scale >= SCALE_LIMIT and y_scale >= SCALE_LIMIT:
        return source_ds, source_gm

    w = np.floor(x_scale * (source_gm.width - 1))
    h = np.floor(y_scale * (source_gm.height - 1))
    downscaled_size = (w if w >= 2 else 2, h if h >= 2 else 2)
    downscale_target_gm = GridMapping.regular(
        size=downscaled_size,
        xy_min=(
            source_gm.xy_bbox[0] + xres_trans / 2,
            source_gm.xy_bbox[1] + yres_trans / 2,
        ),
        xy_res=(xres_trans, yres_trans),
        crs=source_gm.crs,
        tile_size=source_gm.tile_size,
    )
    source_ds = affine_transform_dataset(
        source_ds,
        downscale_target_gm,
        source_gm=source_gm,
        interp_methods=_prep_spatial_interp_methods_downscale(interp_methods),
        agg_methods=agg_methods,
        prevent_nan_propagations=prevent_nan_propagations,
    )
    source_gm = GridMapping.from_dataset(source_ds)

    return source_ds, source_gm


def _reproject_data_array(
    data_array: xr.DataArray,
    var_name: Hashable,
    source_gm: GridMapping,
    target_gm: GridMapping,
    source_xx: da.Array,
    source_yy: da.Array,
    x_coords: da.Array,
    y_coords: da.Array,
    indexing: SourceTileIndexing,
    interp_methods: SpatialInterpMethods | None = None,
    fill_values: FillValues | None = None,
) -> xr.DataArray:
    data_array_expanded = False
    if len(data_array.dims) == 2:
        data_array = data_array.expand_dims({"dummy": 1})
        data_array_expanded = True

    if isinstance(data_array.data, np.ndarray):
        is_numpy_array = True
        array = da.asarray(data_array.data)
    else:
        is_numpy_array = False
        array = data_array.data

    # reorganize data array slice to align with the
    # chunks of source_xx and source_yy
    fill_value = _get_fill_value(fill_values, var_name, data_array)
    interp_method = _get_spatial_interp_method_str(interp_methods, var_name, data_array)
    src_data = _reorganize_tiled_array(array, indexing, fill_value)
    slices_reprojected = []
    # calculate reprojection of each chunk along the 1st (non-spatial) dimension.
    dim0_end = 0
    for chunk_size in array.chunks[0]:
        dim0_start = dim0_end
        dim0_end = dim0_start + chunk_size

        data_reprojected = da.map_blocks(
            _reproject_block,
            source_xx,
            source_yy,
            src_data[dim0_start:dim0_end],
            x_coords,
            y_coords,
            dtype=data_array.dtype,
            chunks=(
                src_data[dim0_start:dim0_end].shape[0],
                source_yy.chunks[0][0],
                source_yy.chunks[1][0],
            ),
            src_x_res=source_gm.x_res,
            src_y_res=source_gm.y_res,
            interp_method=interp_method,
        )
        data_reprojected = data_reprojected[:, : target_gm.height, : target_gm.width]
        slices_reprojected.append(data_reprojected)
    array_reprojected = da.concatenate(slices_reprojected, axis=0)
    if is_numpy_array:
        array_reprojected = array_reprojected.compute()
    if data_array_expanded:
        array_reprojected = array_reprojected[0, :, :]
        dims = (target_gm.xy_dim_names[1], target_gm.xy_dim_names[0])
    else:
        dims = (
            data_array.dims[0],
            target_gm.xy_dim_names[1],
            target_gm.xy_dim_names[0],
        )
    return xr.DataArray(data=array_reprojected, dims=dims, attrs=data_array.attrs)


def _reproject_block(
    source_xx: np.ndarray,
    source_yy: np.ndarray,
    src_data: np.ndarray,
    x_coord: np.ndarray,
    y_coord: np.ndarray,
    src_x_res: int | float,
    src_y_res: int | float,
    interp_method: SpatialInterpMethodStr,
) -> np.ndarray:
    ix = (source_xx - x_coord[0]) / src_x_res
    iy = (source_yy - y_coord[0]) / -src_y_res

    return _sample_array_at_indices(src_data, ix, iy, interp_method)


def _get_src_bboxes_indices(
    transformer: pyproj.Transformer,
    source_gm: GridMapping,
    target_gm: GridMapping,
) -> tuple[SourceTileIndexing, da.Array, da.Array]:
    num_tiles_x = math.ceil(target_gm.width / target_gm.tile_width)
    num_tiles_y = math.ceil(target_gm.height / target_gm.tile_height)

    # get bboxes indices in source grid mapping
    origin = source_gm.x_coords.values[0], source_gm.y_coords.values[0]
    src_ij_bboxes = np.full((4, num_tiles_y, num_tiles_x), -1, dtype=np.int32)
    for idx, xy_bbox in enumerate(target_gm.xy_bboxes):
        j, i = np.unravel_index(idx, (num_tiles_y, num_tiles_x))
        source_xy_bbox = transformer.transform_bounds(*xy_bbox)
        i_min = math.floor((source_xy_bbox[0] - origin[0]) / source_gm.x_res)
        i_max = math.ceil((source_xy_bbox[2] - origin[0]) / source_gm.x_res)
        j_min = math.floor((origin[1] - source_xy_bbox[3]) / source_gm.y_res)
        j_max = math.ceil((origin[1] - source_xy_bbox[1]) / source_gm.y_res)
        src_ij_bboxes[:, j, i] = [i_min, j_min, i_max, j_max]

    # Extend bounding box indices to match the largest bounding box.
    # This ensures uniform chunk sizes, which are required for da.map_blocks.
    i_diff = src_ij_bboxes[2] - src_ij_bboxes[0]
    j_diff = src_ij_bboxes[3] - src_ij_bboxes[1]
    i_diff_max = np.max(i_diff) + 1
    j_diff_max = np.max(j_diff) + 1
    for i in range(num_tiles_x):
        for j in range(num_tiles_y):
            src_ij_bbox = src_ij_bboxes[:, j, i]

            i_half = (i_diff_max - i_diff[j, i]) // 2
            i_start = src_ij_bbox[0] - i_half
            i_end = i_start + i_diff_max

            j_half = (j_diff_max - j_diff[j, i]) // 2
            j_start = src_ij_bbox[1] - j_half
            j_end = j_start + j_diff_max

            src_ij_bboxes[:, j, i] = [i_start, j_start, i_end, j_end]

    # gather the coordinates; coordinates will be extended
    # if they are outside the source grid mapping
    x_coords = np.zeros((i_diff_max, num_tiles_y, num_tiles_x), dtype=np.float32)
    y_coords = np.zeros((j_diff_max, num_tiles_y, num_tiles_x), dtype=np.float32)
    i_min = np.min(src_ij_bboxes[0])
    i_max = np.max(src_ij_bboxes[2])
    j_min = np.min(src_ij_bboxes[[1, 3]])
    j_max = np.max(src_ij_bboxes[[1, 3]])
    x_start = source_gm.x_coords.values[0] + i_min * source_gm.x_res
    x_end = source_gm.x_coords.values[0] + i_max * source_gm.x_res
    x_coord = np.arange(x_start, x_end, source_gm.x_res)
    y_res = source_gm.y_coords.values[1] - source_gm.y_coords.values[0]
    y_start = source_gm.y_coords.values[0] + j_min * y_res
    y_end = source_gm.y_coords.values[0] + j_max * y_res
    y_coord = np.arange(y_start, y_end, y_res)
    for i in range(num_tiles_x):
        for j in range(num_tiles_y):
            src_ij_bbox = src_ij_bboxes[:, j, i]

            i_start = src_ij_bbox[0] - i_min
            i_end = i_start + i_diff_max
            x_coords[:, j, i] = x_coord[i_start:i_end]

            j_start = src_ij_bbox[1] - j_min
            j_end = j_start + j_diff_max
            y_coords[:, j, i] = y_coord[j_start:j_end]

    x_coords = da.from_array(x_coords, chunks=(-1, 1, 1))
    y_coords = da.from_array(y_coords, chunks=(-1, 1, 1))

    pad_width = (
        (
            -min(0, int(j_min)),
            max(0, int(j_max - source_gm.height)),
        ),
        (
            -min(0, int(i_min)),
            max(0, int(i_max - source_gm.width)),
        ),
    )
    src_ij_bboxes[[1, 3]] += pad_width[0][0]
    src_ij_bboxes[[0, 2]] += pad_width[1][0]

    indexing = SourceTileIndexing(
        ij_bboxes=src_ij_bboxes,
        pad_width=pad_width,
        output_size=(y_coords.shape[0] * num_tiles_y, x_coords.shape[0] * num_tiles_x),
        tile_size=(y_coords.shape[0], x_coords.shape[0]),
    )

    return indexing, x_coords, y_coords



# The MIT License (MIT)
# Copyright (c) 2025 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from collections.abc import Callable, Hashable, Iterable, Mapping, Sequence
from dataclasses import dataclass

import dask.array as da
import numpy as np
import pyproj
import xarray as xr

from .constants import (
    AGG_METHODS,
    FILLVALUE_FLOAT,
    FILLVALUE_INT,
    FILLVALUE_UINT8,
    FILLVALUE_UINT16,
    FILLVALUE_UINT32,
    INTERP_METHOD_MAPPING,
    LOG,
    FillValues,
    FloatInt,
    PreventNaNPropagations,
    SpatialAggMethod,
    SpatialAggMethods,
    SpatialInterpMethod,
    SpatialInterpMethodInt,
    SpatialInterpMethods,
    SpatialInterpMethodStr,
)
from .gridmapping import GridMapping

# noinspection PyProtectedMember
from .gridmapping.helpers import _normalize_crs


def get_spatial_coords(ds: xr.Dataset) -> tuple[str, str]:
    """
    Identify the names of horizontal spatial coordinate in an xarray dataset.

    This function checks for standard coordinate name pairs used for horizontal
    spatial referencing: either, ("longitude", "latitude"), ("lon", "lat") or
    ("x", "y"). It returns the detected pair as a tuple in the order (x_coord, y_coord).

    Args:
        ds: The xarray.Dataset to inspect.

    Returns:
        A tuple (x_coord, y_coord) containing the names of the horizontal spatial
        dimensions, e.g., ("longitude", "latitude"), ("lon", "lat") or ("x", "y").

    Raises:
        KeyError: If no recognized spatial dimension pair is found in the dataset.
    """
    if "transformed_x" in ds and "transformed_y" in ds:
        x_coord, y_coord = "transformed_x", "transformed_y"
    elif "latitude" in ds and "longitude" in ds:
        x_coord, y_coord = "longitude", "latitude"
    elif "lat" in ds and "lon" in ds:
        x_coord, y_coord = "lon", "lat"
    elif "y" in ds and "x" in ds:
        x_coord, y_coord = "x", "y"
    else:
        raise KeyError(
            f"No standard spatial coordinates found in dataset. "
            f"Expected pairs ('lon', 'lat') or ('x', 'y'), but found: {list(ds.dims)}."
        )
    return x_coord, y_coord


def get_utm_crs(lon: float, lat: float) -> pyproj.CRS:
    """
    Returns the UTM CRS for a given latitude and longitude as a pyproj.CRS object.

    Args:
        lon: Longitude in decimal degrees.
        lat: Latitude in decimal degrees.

    Returns:
        CRS object corresponding to the correct UTM zone.
    """
    # Standard UTM zone calculation
    zone_number = int((lon + 180) // 6) + 1

    # Norway exception (Western Norway and Svalbard)
    if 56.0 <= lat <= 64.0 and 3.0 <= lon <= 12.0:
        zone_number = 32  # Western Norway
    if 72.0 <= lat <= 84.0:
        # Svalbard zones (special UTM divisions)
        if 0 <= lon < 9:
            zone_number = 31
        elif 9 <= lon < 21:
            zone_number = 33
        elif 21 <= lon < 33:
            zone_number = 35
        elif 33 <= lon < 42:
            zone_number = 37

    # Northern or Southern hemisphere
    epsg = 32600 + zone_number if lat >= 0 else 32700 + zone_number
    return pyproj.CRS.from_epsg(epsg)


def clip_dataset_by_bbox(
    ds: xr.Dataset,
    bbox: Sequence[FloatInt],
    spatial_coords: tuple[str, str] | None = None,
) -> xr.Dataset:
    """
    Clip a xarray Dataset to a specified bounding box.

    This function extracts a spatial subset of the dataset based on the given
    bounding box. It handles both increasing and decreasing orientation of the
    y-axis automatically to ensure correct spatial clipping.

    Args:
        ds: The input xarray.Dataset to be clipped.
        bbox: A sequence of four numbers representing the bounding box in the form
            (min_x, min_y, max_x, max_y).
        spatial_coords: Optional tuple of two spatial coordinate names (x_dim, y_dim),
            e.g., ('lon', 'lat'). If None, the coordinates are inferred automatically.

    Returns:
        A spatial subset of the input dataset clipped to the bounding box.

    Raises:
        ValueError: If `bbox` does not contain exactly four elements.
        KeyError: If spatial coordinate names cannot be determined from the dataset.

    Notes:
        If the bounding box does not overlap with the dataset extent, the returned
        dataset may contain one or more zero-sized dimensions.
    """
    if len(bbox) != 4:
        raise ValueError(f"Expected bbox of length 4, got: {bbox}")

    if spatial_coords is None:
        spatial_coords = get_spatial_coords(ds)
    x_coord, y_coord = spatial_coords

    if ds[x_coord].ndim == 2 and ds[y_coord].ndim == 2:
        ds = _clip_2dcoord_dataset_by_bbox(ds, bbox, x_coord, y_coord)
    elif ds[x_coord].ndim == 1 and ds[y_coord].ndim == 1:
        ds = _clip_1dcoord_dataset_by_bbox(ds, bbox, x_coord, y_coord)
    else:
        raise ValueError(
            f"Unsupported coordinate dimensions: x_coord.ndim={ds[x_coord].ndim}, "
            f"y_coord.ndim={ds[y_coord].ndim}. Expected both 1D or both 2D."
        )

    if any(size == 0 for size in ds.sizes.values()):
        LOG.warning(
            "Clipped dataset contains at least one zero-sized dimension. "
            f"Check if the bounding box {bbox} overlaps with the dataset extent."
        )
    return ds


def _clip_2dcoord_dataset_by_bbox(
    ds: xr.Dataset,
    bbox: Sequence[FloatInt],
    x_coord: str,
    y_coord: str,
) -> xr.Dataset:
    mask = (
        (ds[x_coord] >= bbox[0])
        & (ds[x_coord] <= bbox[2])
        & (ds[y_coord] >= bbox[1])
        & (ds[y_coord] <= bbox[3])
    )
    # Explicitly load the mask into memory here to compute row/column indices using
    # NumPy. This avoids duplicating computations if we stay in Dask for the following
    # operations:
    #   rows = np.any(mask, axis=1)
    #   cols = np.any(mask, axis=0)
    # Note: This will load the entire mask and break chunking, so it is a conscious
    # choice.
    mask = mask.values

    # Find bounding rectangle in index space
    rows = np.any(mask, axis=1)
    cols = np.any(mask, axis=0)
    idxs = np.where(rows)[0]
    if idxs.size == 0:
        row_min, row_max = 0, -1
    else:
        row_min, row_max = idxs[[0, -1]]
    idxs = np.where(cols)[0]
    if idxs.size == 0:
        col_min, col_max = 0, -1
    else:
        col_min, col_max = idxs[[0, -1]]

    y_dim, x_dim = ds[x_coord].dims
    return ds.isel(
        {
            y_dim: slice(row_min, row_max + 1),
            x_dim: slice(col_min, col_max + 1),
        }
    )


def _clip_1dcoord_dataset_by_bbox(
    ds: xr.Dataset,
    bbox: Sequence[FloatInt],
    x_coord: str,
    y_coord: str,
) -> xr.Dataset:
    if ds[y_coord][-1] - ds[y_coord][0] < 0:
        return ds.sel(
            {x_coord: slice(bbox[0], bbox[2]), y_coord: slice(bbox[3], bbox[1])}
        )
    else:
        return ds.sel(
            {x_coord: slice(bbox[0], bbox[2]), y_coord: slice(bbox[1], bbox[3])}
        )


def reproject_bbox(
    source_bbox: Sequence[FloatInt],
    source_crs: pyproj.CRS | str,
    target_crs: pyproj.CRS | str,
) -> Sequence[FloatInt]:
    """Reprojects a bounding box from a source CRS to a target CRS, with optional
    buffering.

    The function transforms a bounding box defined in the source coordinate reference
    system (CRS) to the target CRS using `pyproj`. If the source and target CRS are
    the same, no transformation is performed. An optional buffer (as a fraction of
    width/height) can be applied to expand the resulting bounding box.

    Args:
        source_bbox: The bounding box to reproject, in the form
            (min_x, min_y, max_x, max_y).
        source_crs: The source CRS, as a `pyproj.CRS` or string.
        target_crs: The target CRS, as a `pyproj.CRS` or string.

    Returns:
        A tuple representing the reprojected (and optionally buffered) bounding box:
        (min_x, min_y, max_x, max_y).
    """
    source_crs = _normalize_crs(source_crs)
    target_crs = _normalize_crs(target_crs)
    if source_crs != target_crs:
        t = pyproj.Transformer.from_crs(source_crs, target_crs, always_xy=True)
        target_bbox = t.transform_bounds(*source_bbox, densify_pts=21)
    else:
        target_bbox = source_bbox

    return target_bbox


def bbox_overlap(
    source_bbox: Sequence[FloatInt], target_bbox: Sequence[FloatInt]
) -> float:
    """
    Calculate the fraction of the source bounding box that overlaps with the target
    bounding box relative to the area of the source bounding box.

    Args:
        source_bbox: (min_x, min_y, max_x, max_y)
        target_bbox: (min_x, min_y, max_x, max_y)

    Returns:
        float in [0, 1]
    """
    source_bboxes = _split_bbox_antimeridian(source_bbox)
    target_bboxes = _split_bbox_antimeridian(target_bbox)

    inter_area = 0.0
    for source_bbox in source_bboxes:
        for target_bbox in target_bboxes:
            inter_min_x = max(source_bbox[0], target_bbox[0])
            inter_min_y = max(source_bbox[1], target_bbox[1])
            inter_max_x = min(source_bbox[2], target_bbox[2])
            inter_max_y = min(source_bbox[3], target_bbox[3])

            inter_w = max(0, inter_max_x - inter_min_x)
            inter_h = max(0, inter_max_y - inter_min_y)
            inter_area += inter_w * inter_h
    area_source = 0.0
    for source_bbox in source_bboxes:
        area_source += (source_bbox[2] - source_bbox[0]) * (
            source_bbox[3] - source_bbox[1]
        )

    return inter_area / area_source


def _split_bbox_antimeridian(bbox: Sequence[FloatInt]) -> Sequence[Sequence[FloatInt]]:
    min_x, min_y, max_x, max_y = bbox

    if max_x < min_x:
        return [(min_x, min_y, 180, max_y), (-180, min_y, max_x, max_y)]
    return [bbox]


def resolution_meters_to_degrees(
    resolution: FloatInt | tuple[FloatInt, FloatInt], latitude: FloatInt
) -> tuple[FloatInt, FloatInt]:
    """Convert spatial resolution from meters to degrees in longitude and latitude
    at a given geographic latitude.

    Args:
        resolution: Spatial resolution in meters. Can be a single number
            (applied equally to both axes) or a tuple ``(x_res, y_res)``.
        latitude: Latitude in degrees at which to compute the longitude scaling.

    Returns:
        A tuple `(lon_res_deg, lat_res_deg)` giving the approximate spatial
        resolution in degrees for the latitude and longitude directions.

    Notes:
        - 1 degree of latitude ≈ 111,320 meters (constant approximation).
        - 1 degree of longitude ≈ 111,320 * cos(latitude) meters.

    """
    if not isinstance(resolution, tuple):
        resolution = (resolution, resolution)
    return (
        resolution[0] / (111320 * np.cos(np.deg2rad(latitude))),
        resolution[1] / 111320,
    )


def normalize_grid_mapping(ds: xr.Dataset, gm: GridMapping) -> xr.Dataset:
    """
    Normalize the grid mapping of a dataset to use a standard "spatial_ref" coordinate.

    This function standardizes geospatial metadata by replacing any existing grid
    mapping variable with a unified "spatial_ref" coordinate. It updates the
    "grid_mapping" attribute of all data variables to reference "spatial_ref",
    removes the original grid mapping variable (if present), and adds a new
    "spatial_ref" coordinate with CF-compliant CRS attributes.

    Args:
        ds: The input xarray.Dataset with geospatial grid mapping metadata.
        gm: The GridMapping object associated with the dataset.

    Returns:
        A new dataset with a standardized "spatial_ref" coordinate for grid mapping.
    """
    gm_name = _get_grid_mapping_name(ds)
    if gm_name is not None:
        ds = ds.drop_vars(gm_name)
    ds = ds.assign_coords(spatial_ref=xr.DataArray(0, attrs=gm.crs.to_cf()))
    for var in ds.data_vars:
        ds[var].attrs["grid_mapping"] = "spatial_ref"

    return ds


def _select_variables(
    ds: xr.Dataset, variables: str | Iterable[str] | None = None
) -> xr.Dataset:
    if variables is not None:
        if isinstance(variables, str):
            variables = [variables]
        ds = ds[variables]
    return ds


def _get_grid_mapping_name(ds: xr.Dataset) -> str | None:
    gm_names = []
    for var in ds.data_vars:
        if "grid_mapping" in ds[var].attrs:
            gm_names.append(ds[var].attrs["grid_mapping"])
    if "crs" in ds:
        gm_names.append("crs")
    if "spatial_ref" in ds.coords:
        gm_names.append("spatial_ref")
    gm_names = np.unique(gm_names)
    assert len(gm_names) <= 1, "Multiple grid mapping names found."
    if len(gm_names) == 1:
        return str(gm_names[0])
    else:
        return None


def _can_apply_affine_transform(source_gm: GridMapping, target_gm: GridMapping) -> bool:
    GridMapping.assert_regular(source_gm, name="source_gm")
    GridMapping.assert_regular(target_gm, name="target_gm")
    return _is_equal_crs(source_gm, target_gm)


def _is_equal_crs(source_gm: GridMapping, target_gm: GridMapping) -> bool:
    geographic = source_gm.crs.is_geographic and target_gm.crs.is_geographic
    return geographic or source_gm.crs.equals(target_gm.crs)


def _get_spatial_interp_method(
    interp_methods: SpatialInterpMethods | None,
    key: Hashable,
    var: xr.DataArray,
) -> SpatialInterpMethod:
    def assign_defaults(data_type: np.dtype) -> SpatialInterpMethod:
        if np.issubdtype(data_type, np.bool_):
            return 0
        if np.issubdtype(data_type, np.integer):
            return 0
        return 1

    if isinstance(interp_methods, Mapping):
        interp_method = interp_methods.get(str(key), interp_methods.get(var.dtype))
        if interp_method is None:
            LOG.debug(
                f"Interpolation method could not be derived from the mapping "
                f"`interp_methods` for data variable {key!r} with data type "
                f"{var.dtype!r}. Defaults are assigned."
            )
            interp_method = assign_defaults(var.dtype)
    elif isinstance(interp_methods, int) or isinstance(interp_methods, str):
        interp_method = interp_methods
    else:
        interp_method = assign_defaults(var.dtype)

    return interp_method


def _get_spatial_interp_method_int(
    interp_methods: SpatialInterpMethods | None,
    key: Hashable,
    var: xr.DataArray,
) -> SpatialInterpMethodInt:
    interp_method = _get_spatial_interp_method(interp_methods, key, var)
    if isinstance(interp_method, str):
        interp_method = INTERP_METHOD_MAPPING[interp_method]
    return interp_method


def _get_spatial_interp_method_str(
    interp_methods: SpatialInterpMethods | None,
    key: Hashable,
    var: xr.DataArray,
) -> SpatialInterpMethodStr:
    interp_method = _get_spatial_interp_method(interp_methods, key, var)
    if isinstance(interp_method, int):
        interp_method = INTERP_METHOD_MAPPING[interp_method]
    return interp_method


def _prep_spatial_interp_methods_downscale(
    interp_methods: SpatialInterpMethods | None,
) -> SpatialInterpMethods | None:
    if interp_methods == "triangular":
        return "bilinear"
    elif (
        isinstance(interp_methods, Mapping) and "triangular" in interp_methods.values()
    ):
        return {
            k: ("bilinear" if v == "triangular" else v)
            for k, v in interp_methods.items()
        }
    return interp_methods


def _get_spatial_interp_methods(
    ds: xr.Dataset, interp_methods: SpatialInterpMethods | None
) -> Mapping[Hashable, SpatialInterpMethod]:
    interp_methods_new = {}
    for name, var in ds.data_vars.items():
        interp_methods_new[name] = _get_spatial_interp_method(interp_methods, name, var)
    return interp_methods_new


def _get_spatial_agg_method(
    agg_methods: SpatialAggMethods | None,
    key: Hashable,
    var: xr.DataArray,
) -> Callable:
    def assign_defaults(data_type: np.dtype) -> SpatialAggMethod:
        if np.issubdtype(data_type, np.bool_):
            return "center"
        if np.issubdtype(data_type, np.integer):
            return "center"
        return "mean"

    if isinstance(agg_methods, Mapping):
        agg_method = agg_methods.get(str(key), agg_methods.get(var.dtype))
        if agg_method is None:
            LOG.debug(
                f"Aggregation method could not be derived from the mapping `agg_methods` "
                f"for data variable {key!r} with data type {var.dtype!r}. Defaults "
                f"are assigned."
            )
            agg_method = assign_defaults(var.dtype)
    elif isinstance(agg_methods, str):
        agg_method = agg_methods
    else:
        agg_method = assign_defaults(var.dtype)

    return AGG_METHODS[agg_method]


def _get_prevent_nan_propagation(
    prevent_nan_propagations: PreventNaNPropagations | None,
    key: Hashable,
    var: xr.DataArray,
) -> bool:
    if isinstance(prevent_nan_propagations, Mapping):
        prevent_nan_propagation = prevent_nan_propagations.get(
            str(key), prevent_nan_propagations.get(var.dtype)
        )
        if prevent_nan_propagation is None:
            LOG.debug(
                f"The method to prevent NaN propagation could not be derived from "
                f"the mapping `prevent_nan_propagations`  for data variable {key!r} "
                f"with data type {var.dtype!r}. Defaults are assigned."
            )
            prevent_nan_propagation = False
    elif isinstance(prevent_nan_propagations, bool):
        prevent_nan_propagation = prevent_nan_propagations
    else:
        prevent_nan_propagation = False

    return prevent_nan_propagation


def _get_fill_value(
    fill_values: int | float | Mapping[np.dtype | str, int | float] | None,
    key: Hashable,
    var: xr.DataArray,
) -> int:
    # noinspection PyShadowingNames
    def assign_defaults(data_type: np.dtype) -> int:
        if data_type == np.uint8:
            fill_value = FILLVALUE_UINT8
        elif data_type == np.uint16:
            fill_value = FILLVALUE_UINT16
        elif data_type == np.uint32:
            fill_value = FILLVALUE_UINT32
        elif np.issubdtype(data_type, np.integer):
            fill_value = FILLVALUE_INT
        elif np.issubdtype(data_type, np.bool_):
            fill_value = 0
        else:
            fill_value = FILLVALUE_FLOAT
        return fill_value

    if isinstance(fill_values, Mapping):
        fill_value = fill_values.get(str(key), fill_values.get(var.dtype))
        if fill_value is None:
            LOG.debug(
                f"Fill value could not be derived from the mapping `fill_values` "
                f"for data variable {key!r} with data type {var.dtype!r}. Defaults "
                f"are assigned."
            )
            fill_value = assign_defaults(var.dtype)
    elif fill_values is not None:
        fill_value = fill_values
    else:
        fill_value = assign_defaults(var.dtype)

    return fill_value


def _sample_array_at_indices(
    data: np.ndarray,
    ix: np.ndarray,
    iy: np.ndarray,
    interp_method: SpatialInterpMethodStr,
) -> np.ndarray:
    """
    Sample a 3d array at fractional indices (iy, ix).
    """
    if interp_method == "nearest":
        ix_i = np.ceil(ix - 0.5).astype(np.intp)
        iy_i = np.ceil(iy - 0.5).astype(np.intp)
        return data[:, iy_i, ix_i]

    ix_floor = np.floor(ix).astype(np.intp)
    iy_floor = np.floor(iy).astype(np.intp)
    ix_ceil = np.ceil(ix).astype(np.intp)
    iy_ceil = np.ceil(iy).astype(np.intp)

    dx = ix - ix_floor
    dy = iy - iy_floor

    v00 = data[:, iy_floor, ix_floor]
    v01 = data[:, iy_floor, ix_ceil]
    v10 = data[:, iy_ceil, ix_floor]
    v11 = data[:, iy_ceil, ix_ceil]

    if interp_method == "bilinear":
        u0 = v00 + dx * (v01 - v00)
        u1 = v10 + dx * (v11 - v10)
        return u0 + dy * (u1 - u0)

    if interp_method == "triangular":
        dx3 = dx[np.newaxis, ...]
        dy3 = dy[np.newaxis, ...]

        tri1 = v00 + dx3 * (v01 - v00) + dy3 * (v10 - v00)
        tri2 = v11 + (1 - dx3) * (v10 - v11) + (1 - dy3) * (v01 - v11)
        mask3 = (dx + dy < 1.0)[np.newaxis, ...]
        return np.where(mask3, tri1, tri2)

    raise NotImplementedError(
        f"interp_methods must be one of 0, 1, 'nearest', 'bilinear', 'triangular', "
        f"was '{interp_method}'."
    )


def _create_empty_dataset(
    source_ds: xr.Dataset,
    source_gm: GridMapping,
    target_gm: GridMapping,
    fill_values: FillValues | None = None,
) -> xr.Dataset:
    x_name, y_name = source_gm.xy_var_names
    coords = source_ds.coords.to_dataset()
    coords = coords.drop_vars((x_name, y_name), errors="ignore")
    x_name, y_name = target_gm.xy_var_names
    coords[x_name] = target_gm.x_coords
    coords[y_name] = target_gm.y_coords
    coords["spatial_ref"] = xr.DataArray(0, attrs=target_gm.crs.to_cf())
    target_ds = xr.Dataset(coords=coords, attrs=source_ds.attrs)
    for key, data in source_ds.data_vars.items():
        shape = list(source_ds[key].shape)
        shape[-1] = target_gm.width
        shape[-2] = target_gm.height
        dims = list(source_ds[key].dims)
        dims[-1] = target_gm.xy_var_names[0]
        dims[-2] = target_gm.xy_var_names[1]
        if source_ds[key].ndim == 3:
            chunks = (
                source_ds[key].chunks[0][0],
                target_gm.height,
                target_gm.width,
            )
        else:
            chunks = (target_gm.height, target_gm.width)
        target_ds[key] = xr.DataArray(
            da.full(
                shape,
                fill_value=_get_fill_value(fill_values, key, data),
                chunks=chunks,
            ),
            dims=dims,
            attrs=source_ds[key].attrs,
        )
    return target_ds


def _clip_if_needed(
    source_ds: xr.Dataset,
    source_gm: GridMapping,
    target_gm: GridMapping,
    fill_values: FillValues | None,
    transformer: pyproj.Transformer | None = None,
) -> tuple[xr.Dataset, GridMapping, bool]:
    """
    Clip the source dataset to the target bounding box if overlap is limited.
    """
    if transformer is None:
        target_bbox = target_gm.xy_bbox
        buffer_x = 2 * max(source_gm.x_res, target_gm.x_res)
        buffer_y = 2 * max(source_gm.y_res, target_gm.y_res)
    else:
        target_bbox = transformer.transform_bounds(*target_gm.xy_bbox)
        target_x_res = (target_bbox[2] - target_bbox[0]) / target_gm.width
        target_y_res = (target_bbox[3] - target_bbox[1]) / target_gm.height
        buffer_x = 2 * max(source_gm.x_res, target_x_res)
        buffer_y = 2 * max(source_gm.y_res, target_y_res)

    overlap = bbox_overlap(source_gm.xy_bbox, target_bbox)
    if overlap == 0.0:
        LOG.info(
            "Target grid mapping does not overlap with the source grid mapping. "
            "Returning empty target dataset."
        )
        empty_ds = _create_empty_dataset(source_ds, source_gm, target_gm, fill_values)
        return empty_ds, target_gm, True
    if overlap >= 0.8:
        return source_ds, source_gm, False

    # Expand target bbox slightly to ensure coverage
    bbox = [
        target_bbox[0] - buffer_x,
        target_bbox[1] - buffer_y,
        target_bbox[2] + buffer_x,
        target_bbox[3] + buffer_y,
    ]
    clipped = clip_dataset_by_bbox(source_ds, bbox)

    if any(clipped.sizes[source_gm.xy_dim_names[i]] < 2 for i in range(2)):
        LOG.warning(
            "Clipped dataset contains a spatial dimension with size < 2. "
            "Returning empty target dataset."
        )
        empty_ds = _create_empty_dataset(source_ds, source_gm, target_gm, fill_values)
        return empty_ds, target_gm, True

    return clipped, GridMapping.from_dataset(clipped), False


@dataclass(frozen=True)
class SourceTileIndexing:
    ij_bboxes: np.ndarray  # (4, ny_tiles, nx_tiles)
    pad_width: tuple[tuple[int, int], tuple[int, int]]
    output_size: tuple[int, int]
    tile_size: tuple[int, int]


def _reorganize_tiled_array(
    array: da.Array,
    indexing: SourceTileIndexing,
    fill_value: FloatInt,
) -> da.Array:
    """
    Reorganize a 2D or 3D array into uniformly sized tiles based on source index bboxes.
    """

    is_3d = array.ndim == 3
    if not is_3d:
        array = array[None, ...]

    pad_width = ((0, 0),) + indexing.pad_width
    padded = da.pad(array, pad_width, mode="constant", constant_values=fill_value)

    out_shape = (
        array.shape[0],
        indexing.output_size[0],
        indexing.output_size[1],
    )
    chunks = (
        array.chunks[0],
        indexing.tile_size[0],
        indexing.tile_size[1],
    )

    out = da.zeros(out_shape, chunks=chunks, dtype=array.dtype)

    ny, nx = indexing.ij_bboxes.shape[1:]
    th, tw = indexing.tile_size

    for j in range(ny):
        for i in range(nx):
            bbox = indexing.ij_bboxes[:, j, i]

            y0, y1 = j * th, (j + 1) * th
            x0, x1 = i * tw, (i + 1) * tw

            if bbox[0] == -1:
                out[:, y0:y1, x0:x1] = da.full(
                    (array.shape[0], th, tw),
                    fill_value,
                    chunks=chunks,
                    dtype=array.dtype,
                )
            else:
                out[:, y0:y1, x0:x1] = padded[
                    :,
                    bbox[1] : bbox[3],
                    bbox[0] : bbox[2],
                ]

    return out if is_3d else out[0]


def _map_to_source_indices(pixel_target_ij, indexing):
    """Convert tile-local fractional source pixel indices to original source indices.

    Applies per-tile source index offsets and removes padding introduced during
    tiling, returning fractional (ix, iy) indices in the original source grid.
    """
    offsets_x = da.zeros_like(pixel_target_ij[0])
    offsets_y = da.zeros_like(pixel_target_ij[1])

    ny, nx = indexing.ij_bboxes.shape[1:]
    th, tw = indexing.tile_size

    for j in range(ny):
        for i in range(nx):
            bbox = indexing.ij_bboxes[:, j, i]
            offsets_x[th * j : th * (j + 1), tw * i : tw * (i + 1)] = bbox[0]
            offsets_y[th * j : th * (j + 1), tw * i : tw * (i + 1)] = bbox[1]

    offsets_x -= indexing.pad_width[1][0]
    offsets_y -= indexing.pad_width[0][0]

    return da.stack([pixel_target_ij[0] + offsets_x, pixel_target_ij[1] + offsets_y])
